# -*- coding: utf-8 -*-
# ===========================================================================
# General toolkits for Signal processing in pure python, numpy and scipy
# The code is collected and modified from following library:
# * https://gist.github.com/kastnerkyle/179d6e9a88202ab0a2fe
#       License: BSD 3-clause
#       Authors: Kyle Kastner
#       Harvest, Cheaptrick, D4C, WORLD routines based on MATLAB code from M. Morise
#       http://ml.cs.yamanashi.ac.jp/world/english/
#       MGC code based on r9y9 (Ryusshichi Yamamoto) MelGeneralizedCepstrums.jl
#       Pieces also adapted from SPTK
# * https://github.com/librosa/librosa
#       License: https://github.com/librosa/librosa/blob/master/LICENSE.md
# * http://www-lium.univ-lemans.fr/sidekit/
#       Authors: Anthony Larcher & Kong Aik Lee & Sylvain Meignier
#       License: GNU Library or Lesser General Public License (LGPL)
# ===========================================================================
from __future__ import print_function, division, absolute_import

import os
import six
import copy
import warnings
import subprocess
from io import BytesIO
from numbers import Number
from six import string_types

import numpy as np
import scipy as sp
from scipy import linalg, fftpack, signal
from numpy.lib.stride_tricks import as_strided
try:
  from odin.utils import cache_memory, cache_disk
except ImportError:
  def cache_memory(func):
    return func

  def cache_disk(func):
    return func

# Constrain STFT block sizes to 512 KB
MAX_MEM_BLOCK = 2**8 * 2**11
# ===========================================================================
# Helper
# ===========================================================================
def anything2wav(inpath, outpath=None,
                 channel=None, sample_rate=None, codec=None,
                 start=None, end=None,
                 dataset='unknown', return_data=True):
  """

  Parameters
  ----------
  inpath : string
    path to audio file

  outpath : {string, None}
    if None, return `BufferReader` contain the result of the conversion,
    call `read()` to get all the binary data of the .wav file.
    Otherwise, return the command for execute or the path of output file

  channel : {None, 0, 1}
    specific channel for the conversion
    0 for left channel, 1 for right channel

  sample_rate : {None, integer}
    sample rate for down-sampling

  dataset : {'unknown', 'voxceleb2'}
    predefine recipe for some datasets, supported datasets include:
     - voxceleb2
     - mx6 (mixer 6)

  codec : {None, string}
    the desire codec for output file
    for `sox`:
      'signed-integer' - pcm16
      'unsigned-integer' - pcm8
      'floating-point' -
      'a-law' - PCM A-law
      'u-law' -
      'mu-law' - PCM mu-law
      'oki-adpcm' -
    'ima-adpcm', 'ms-adpcm', 'gsm-full-rate'
    for `ffmpeg`:
      'alaw ' - PCM A-law
      'f32be' - PCM 32-bit floating-point big-endian
      'f32le' - PCM 32-bit floating-point little-endian
      'f64be' - PCM 64-bit floating-point big-endian
      'f64le' - PCM 64-bit floating-point little-endian
      'mulaw' - PCM mu-law
      's16be' - PCM signed 16-bit big-endian
      's16le' - PCM signed 16-bit little-endian (pcm 16)
      's24be' - PCM signed 24-bit big-endian
      's24le' - PCM signed 24-bit little-endian
      's32be' - PCM signed 32-bit big-endian
      's32le' - PCM signed 32-bit little-endian
      's8   ' - PCM signed 8-bit
      'u16be' - PCM unsigned 16-bit big-endian
      'u16le' - PCM unsigned 16-bit little-endian
      'u24be' - PCM unsigned 24-bit big-endian
      'u24le' - PCM unsigned 24-bit little-endian
      'u32be' - PCM unsigned 32-bit big-endian
      'u32le' - PCM unsigned 32-bit little-endian
      'u8   ' - PCM unsigned 8-bit (pcm8)

  start : {None, float}
    start time in second for trimming

  end : {None, float}
    end time in second (relative to the beginning) for trimming

  return_data : bool (default: True)
    if False, return the output path of .wav or the shell command in case
    `outpath=None`,
    otherwise, execute the command using `subprocess.Popen` and
    return the tuple of (data, sample_rate)

  Return
  ------
  if `return_data=True`: tuple(y, sr)
    y : numpy.ndarray (n_samples, n_channels)
    sr : integer (sample rate)
  else: string
    shell command for execute the conversion

  Note
  ----
  by default: we only accept channel of 0 (left) or 1 (right),
  and this value will be automatically converted to appropriate
  value for each toolkit:
    - for `sox` and `sph2pipe` the channel is 1 or 2
    - for `ffmpeg` the channel is 0 or 1

  """
  inpath = str(inpath)
  if dataset is None:
    dataset = 'unknown'
  dataset = str(dataset).lower()
  ext = os.path.splitext(inpath)[-1]
  is_given_outpath = True
  if outpath is None:
    outpath = '-'
    is_given_outpath = False
  # ====== unknown start and end time ====== #
  if start == '-':
    start = None
  if end == '-':
    end = None
  # ====== init ====== #
  tool = None
  cmd = None
  options = []
  # ====== mixer6 dataset ====== #
  if dataset.lower() == 'mx6':
    if '/ulaw_sphere/' in inpath:
      cmd = 'sph2pipe -f wav -p -c %d %s %s' % \
      (int(channel) + 1, inpath, '' if not is_given_outpath else outpath)
    else:
      # codec
      if codec is not None:
        codec = '-e signed-integer' if 'pcm16' in codec else '-e %s' % str(codec)
      else:
        codec = ''
      # main command
      cmd = 'sox -t flac %s %s -t wav %s %s trim %s =%s' % \
      (inpath,
       '' if sample_rate is None else '-r %d' % int(sample_rate),
       codec,
       outpath,
       start, end)
  # ====== other using SOX ====== #
  elif ext.lower() in ('.sph', '.wav', '.flac'):
    tool = 'sox'
    # options.append('-G')
    options.append('--no-dither')
    options.append('-t %s' % ext[1:])
    options.append(str(inpath))
    # downsample
    options.append('-r %d' % int(sample_rate)
                   if sample_rate is not None else '')
    options.append('-t wav')
    # codec
    if codec is not None:
      codec = 'signed-integer' if 'pcm16' in codec else str(codec)
    else:
      codec = ''
    if len(codec) > 0:
      options.append('-e %s' % codec)
    # output path
    options.append(str(outpath))
    # channel selection
    if channel is not None: # for sox channel is 1 or 2
      options.append('remix %d' % (int(channel) + 1))
    # trim audio
    if start is not None or end is not None:
      options.append('trim %s %s' %
                     ('%f' % start if start is not None else '',
                      '=%f' % end if end is not None else ''))
  # ====== other using ffmpeg ====== #
  elif ext.lower() in ('.m4a', '.mp3'):
    tool = 'ffmpeg'
    # force overwrite
    options.append('-y')
    # log level
    options.append('-v 8')
    options.append('-i %s' % inpath)
    # channel selection
    if channel is not None:
      options.append('-map_channel 0.0.%d' % int(channel))
    # downsample
    options.append('-ar %d' % int(sample_rate)
                   if sample_rate is not None else '')
    # output wav
    options.append('-f wav')
    # codec
    if dataset == 'voxceleb2':
      options.append('-acodec pcm_s16le')
    else:
      if codec is None:
        codec = ''
      else:
        codec = 'pcm_s16le' if 'pcm16' in codec else str(codec)
      if len(codec) > 0:
        options.append('-acodec %s' % codec)
    # output path
    options.append(str(outpath))
    # trim audio
    if start is not None:
      options.append('-ss %f' % start)
    if end is not None:
      options.append('-to %f' % end)
  else:
    raise ValueError("Nonsupport for file extension: %s" % inpath)
  # ====== merge everything ====== #
  if cmd is None:
    cmd = tool + ' ' + ' '.join([i for i in options
                                 if len(i) > 0])
  # ====== only return the shell command ====== #
  if not return_data:
    return cmd
  # ====== run the command and return the data ====== #
  import soundfile
  # read bytes data from PIPE output
  if outpath in ('-', '- |', '-|'):
    with subprocess.Popen(cmd, shell=True,
                          stdout=subprocess.PIPE,
                          stderr=subprocess.PIPE) as task:
      data = BytesIO(task.stdout.read())
      try:
        data = soundfile.read(data) # y, sr
      except Exception as e:
        print("Error running command:", cmd)
        print("Stdout:", str(task.stdout.read(), 'utf-8'))
        print("Stderr:", str(task.stderr.read(), 'utf-8'))
        raise e
      return data
  # read audio file from outpath
  else:
    try:
      with subprocess.Popen(cmd, shell=True) as task:
        pass
    except Exception as e:
      print("Error running command:", cmd)
      print("Stdout:", str(task.stdout.read(), 'utf-8'))
      print("Stderr:", str(task.stderr.read(), 'utf-8'))
      raise e
    with open(outpath, 'rb') as f:
      return soundfile.read(f)

# ===========================================================================
# VAD
# ===========================================================================
VAD_MODE_STRICT = 1.2
VAD_MODE_STANDARD = 2.
VAD_MODE_SENSITIVE = 2.4
__current_vad_mode = VAD_MODE_STANDARD # alpha for vad energy

def set_vad_mode(mode):
  """
  Paramters
  ---------
  mode: float
      a number from 1.0 to 2.4, the higher the number, the more
      sensitive it is to any high-energy segments.
  """
  if isinstance(mode, Number):
    global __current_vad_mode
    mode = min(max(mode, 1.), 2.4)
    __current_vad_mode = float(mode)

def vad_energy(log_energy, distrib_nb=3, nb_train_it=25):
  """ Fitting Gaussian mixture model on the log-energy and the voice
  activity is the component with highest energy.

  Return
  ------
  vad: array of 0, 1
  threshold: scalar
  """
  from sklearn.exceptions import ConvergenceWarning
  from sklearn.mixture import GaussianMixture
  # center and normalize the energy
  log_energy = (log_energy - np.mean(log_energy)) / np.std(log_energy)
  if log_energy.ndim == 1:
    log_energy = log_energy[:, np.newaxis]
  # create mixture model: diag, spherical
  world = GaussianMixture(
      n_components=distrib_nb, covariance_type='diag',
      init_params='kmeans', max_iter=nb_train_it,
      weights_init=np.ones(distrib_nb) / distrib_nb,
      means_init=(-2 + 4.0 * np.arange(distrib_nb) / (distrib_nb - 1))[:, np.newaxis],
      precisions_init=np.ones((distrib_nb, 1)),
  )
  try:
    with warnings.catch_warnings():
      warnings.filterwarnings("ignore", category=ConvergenceWarning)
      world.fit(log_energy)
  except (ValueError, IndexError): # index error because of float32 cumsum
    if distrib_nb - 1 >= 2:
      return vad_energy(log_energy,
                        distrib_nb=distrib_nb - 1,
                        nb_train_it=nb_train_it)
    return np.zeros(shape=(log_energy.shape[0],)), 0
  # Compute threshold
  threshold = world.means_.max() - \
      __current_vad_mode * np.sqrt(1.0 / world.precisions_[world.means_.argmax(), 0])
  # Apply frame selection with the current threshold
  label = log_energy.ravel() > threshold
  return label, threshold

def vad_threshold(frames, threshold=35):
  """
  threshold : scalar (30,40)
  """
  energies = 20 * np.log10(np.std(frames, axis=0) + np.finfo(float).eps)
  max_energy = np.max(energies)
  return (energies > max_energy - threshold) & (energies > -55)

def vad_split_audio(s, sr, maximum_duration=30, minimum_duration=None,
                    frame_length=128, nb_mixtures=3, threshold=0.6,
                    return_vad=False, return_voices=False, return_cut=False):
  """ Splitting an audio based on VAD indicator.
  * The audio is segmented into multiple with length given by `frame_length`
  * Log-energy is calculated for each frames
  * Gaussian mixtures with `nb_mixtures` is fitted, and output vad indicator
    for each frames.
  * A flat window (ones-window) of `frame_length` is convolved with the
    vad indices.
  * All frames within the percentile >= `threshold` is treated as voiced.
  * The splitting process is greedy, frames is grouped until reaching the
    maximum duration

  Parameters
  ----------
  s: 1-D numpy.ndarray
      loaded audio array
  sr: int
      sample rate
  maximum_duration: float (second)
      maximum duration of each segments in seconds
  minimum_duration: None, or float (second)
      all segments below this length will be merged into longer segments,
      if None, any segments with half of the `maximum_duration`
      are considered.
  frame_length: int
      number of frames for windowing
  nb_mixtures: int
      number of Gaussian mixture for energy-based VAD (the higher
      the more overfitting).
  threshold: float (0. to 1.)
      The higher the values, the more frames are considered as voiced,
      this value is the lower percentile of voiced frames.
  return_vad: bool
      if True, return VAD confident values
  return_voices: bool
      if True, return the voices frames indices
  return_cut: bool
      if True, return the cut points of the audio.

  Return
  ------
  segments: list of audio arrays
  vad (optional): list of 0, 1 for VAD indices
  voices (optional): list of thresholded VAD for more precise voices frames.
  cut (optional): list of indicator 0, 1 (1 for the cut point)

  Note
  ----
  this function does not guarantee the output audio length is always
  smaller than `maximum_duration`, the higher the threshold, the better
  chance you get everything smaller than `maximum_duration`
  """
  frame_length = int(frame_length)
  maximum_duration = maximum_duration * sr
  results = []
  # ====== check if audio long enough ====== #
  if len(s) < maximum_duration:
    if return_cut or return_vad or return_voices:
      raise ValueError("Cannot return `cut` points, `vad` or `voices` since"
                  "the original audio is shorter than `maximum_duration`, "
                  "hence, no need for splitting.")
    return [s]
  maximum_duration /= frame_length
  if minimum_duration is None:
    minimum_duration = maximum_duration // 2
  else:
    minimum_duration = minimum_duration * sr / frame_length
    minimum_duration = np.clip(minimum_duration, 0., 0.99 * maximum_duration)
  # ====== start spliting ====== #
  frames = segment_axis(s, frame_length, frame_length,
                        axis=0, end='pad', pad_value=0.)
  energy = get_energy(frames, log=True)
  vad = vad_energy(energy, distrib_nb=nb_mixtures, nb_train_it=33)[0]
  vad = smooth(vad, win=frame_length, window='flat')
  # explicitly return VAD
  if return_vad:
    results.append(vad)
  # ====== get all possible sliences ====== #
  # all voice indices
  indices = np.where(vad >= np.percentile(vad, q=threshold * 100))[0].tolist()
  if len(vad) - 1 not in indices:
    indices.append(len(vad) - 1)
  # explicitly return voiced frames
  if return_voices:
    tmp = np.zeros(shape=(len(vad),))
    tmp[indices] = 1
    results.append(tmp)
  # ====== spliting the audio ====== #
  segments = []
  start = 0
  prev_end = 0
  # greedy adding new frames to reach desire maximum length
  for end in indices:
    # over-reach the maximum length
    if end - start > maximum_duration:
      segments.append((start, prev_end))
      start = prev_end
    # exact maximum length
    elif end - start == maximum_duration:
      segments.append((start, end))
      start = end
    prev_end = end
  # still no segments found
  if len(segments) == 0:
    segments = [(indices[0], indices[-1])]
  # add ending index if necessary
  if indices[-1] != segments[-1][-1]:
    segments.append((start, indices[-1]))
  # re-fining, short segments will be merged into bigger onces
  found_under_length = True
  while found_under_length and len(segments) > 1:
    new_segments = []
    found_under_length = False
    for (s1, e1), (s2, e2) in zip(segments, segments[1:]):
      if (e1 - s1) < minimum_duration or (e2 - s2) < minimum_duration:
        new_segments.append((s1, e2))
        found_under_length = True
      # keep both of the segments
      else:
        new_segments.append((s1, e1))
        new_segments.append((s2, e2))
    segments = new_segments
  # explicitly return cut points
  if return_cut:
    tmp = np.zeros(shape=(segments[-1][-1] + 1,))
    for i, j in segments:
      tmp[i] = 1; tmp[j] = 1
    results.append(tmp)
  # ====== convert everythng to raw signal index ====== #
  segments = [[i * frame_length, j * frame_length]
              for i, j in segments]
  segments[-1][-1] = s.shape[0]
  # cut segments out of raw audio array
  segments = [s[i:j] for i, j in segments]
  results = [segments] + results
  return results[0] if len(results) == 1 else results

# ===========================================================================
#
# ===========================================================================
def loudness2intensity(loudness):
  """ 60dB is consider the standard """
  if loudness.ndim == 2:
    loudness = loudness[:, 0]
  return loudness * 60

def hz2mel(frequencies):
  """Convert Hz to Mels
  Original code: librosa

  Examples
  --------
  >>> hz2mel(60)
  array([ 0.9])
  >>> hz2mel([110, 220, 440])
  array([ 1.65,  3.3 ,  6.6 ])

  Parameters
  ----------
  frequencies   : np.ndarray [shape=(n,)] , float
      scalar or array of frequencies

  Returns
  -------
  mels        : np.ndarray [shape=(n,)]
      input frequencies in Mels

  See Also
  --------
  mel_to_hz
  """
  frequencies = np.atleast_1d(frequencies)
  # Fill in the linear part
  f_min = 0.0
  f_sp = 200.0 / 3
  mels = (frequencies - f_min) / f_sp
  # Fill in the log-scale part
  min_log_hz = 1000.0                         # beginning of log region (Hz)
  min_log_mel = (min_log_hz - f_min) / f_sp   # same (Mels)
  logstep = np.log(6.4) / 27.0                # step size for log region

  log_t = (frequencies >= min_log_hz)
  mels[log_t] = min_log_mel + np.log(frequencies[log_t] / min_log_hz) / logstep
  return mels

def mel2hz(mels):
  """Convert mel bin numbers to frequencies
  Original code: librosa

  Examples
  --------
  >>> mel2hz(3)
  array([ 200.])
  >>> mel2hz([1,2,3,4,5])
  array([  66.667,  133.333,  200.   ,  266.667,  333.333])

  Parameters
  ----------
  mels          : np.ndarray [shape=(n,)], float
      mel bins to convert

  Returns
  -------
  frequencies   : np.ndarray [shape=(n,)]
      input mels in Hz

  See Also
  --------
  hz_to_mel
  """

  mels = np.atleast_1d(mels)

  # Fill in the linear scale
  f_min = 0.0
  f_sp = 200.0 / 3
  freqs = f_min + f_sp * mels

  # And now the nonlinear scale
  min_log_hz = 1000.0                         # beginning of log region (Hz)
  min_log_mel = (min_log_hz - f_min) / f_sp   # same (Mels)
  logstep = np.log(6.4) / 27.0                # step size for log region
  log_t = (mels >= min_log_mel)

  freqs[log_t] = min_log_hz * np.exp(logstep * (mels[log_t] - min_log_mel))
  return freqs

def mel_frequencies(n_mels=128, fmin=0.0, fmax=11025.0):
  """Compute the center frequencies of mel bands.
  Original code: librosa

  Parameters
  ----------
  n_mels    : int > 0 [scalar]
      number of Mel bins
  fmin      : float >= 0 [scalar]
      minimum frequency (Hz)
  fmax      : float >= 0 [scalar]
      maximum frequency (Hz)

  Returns
  -------
  bin_frequencies : ndarray [shape=(n_mels,)]
      vector of n_mels frequencies in Hz which are uniformly spaced on the Mel
      axis.

  Examples
  --------
  >>> librosa.mel_frequencies(n_mels=40)
  array([     0.   ,     85.317,    170.635,    255.952,
            341.269,    426.586,    511.904,    597.221,
            682.538,    767.855,    853.173,    938.49 ,
           1024.856,   1119.114,   1222.042,   1334.436,
           1457.167,   1591.187,   1737.532,   1897.337,
           2071.84 ,   2262.393,   2470.47 ,   2697.686,
           2945.799,   3216.731,   3512.582,   3835.643,
           4188.417,   4573.636,   4994.285,   5453.621,
           5955.205,   6502.92 ,   7101.009,   7754.107,
           8467.272,   9246.028,  10096.408,  11025.   ])

  """
  min_mel = hz2mel(fmin)
  max_mel = hz2mel(fmax)
  mels = np.linspace(min_mel, max_mel, n_mels)
  return mel2hz(mels)

def db2power(S_db, ref=1.0):
  '''Convert a dB-scale spectrogram to a power spectrogram.

  This effectively inverts `power_to_db`:

      `db_to_power(S_db) ~= ref * 10.0**(S_db / 10)`

  Original code: librosa

  Parameters
  ----------
  S_db : np.ndarray
      dB-scaled spectrogram
  ref : number > 0
      Reference power: output will be scaled by this value

  Returns
  -------
  S : np.ndarray
      Power spectrogram

  Notes
  -----
  This function caches at level 30.
  '''
  return ref * np.power(10.0, 0.1 * S_db)

def power2db(S, ref=1.0, amin=1e-10, top_db=80.0):
  """Convert a power spectrogram (amplitude/magnitude squared)
  to decibel (dB) units (using logarithm)

  This computes the scaling ``10 * log10(S / ref)`` in a numerically
  stable way.

  Original code: librosa

  Parameters
  ----------
  S : np.ndarray
      input power
  ref : scalar or call-able
      If scalar, the amplitude `abs(S)` is scaled relative to `ref`:
      `10 * log10(S / ref)`.
      Zeros in the output correspond to positions where `S == ref`.
      If call-able, the reference value is computed as `ref(S)`.
  amin : float > 0 [scalar]
      minimum threshold for `abs(S)` and `ref`
  top_db : float >= 0 [scalar]
      threshold the output at `top_db` below the peak:
      ``max(10 * log10(S)) - top_db``

  Returns
  -------
  S_db   : np.ndarray
      ``S_db ~= 10 * log10(S) - 10 * log10(ref)``
  """
  if amin <= 0:
    raise ValueError('amin must be strictly positive')
  magnitude = np.abs(S)
  if hasattr(ref, '__call__'):
    # User supplied a function to calculate reference power
    ref_value = ref(magnitude)
  else:
    ref_value = np.abs(ref)
  log_spec = 10.0 * np.log10(np.maximum(amin, magnitude))
  log_spec -= 10.0 * np.log10(np.maximum(amin, ref_value))
  # clip top db
  if top_db is not None:
    if top_db < 0:
      raise ValueError('top_db must be non-negative')
    log_spec = np.maximum(log_spec, log_spec.max() - top_db)
  return log_spec

@cache_memory
def dct_filters(n_filters, n_input):
  """Discrete cosine transform (DCT type-III) basis.

  .. [1] http://en.wikipedia.org/wiki/Discrete_cosine_transform

  Original code: librosa

  Parameters
  ----------
  n_filters : int > 0 [scalar]
      number of output components (DCT filters)

  n_input : int > 0 [scalar]
      number of input components (frequency bins)

  Returns
  -------
  dct_basis: np.ndarray [shape=(n_filters, n_input)]
      DCT (type-III) basis vectors [1]_

  Notes
  -----
  This function caches at level 10.

  Examples
  --------
  >>> n_fft = 2048
  >>> dct_filters = librosa.filters.dct(13, 1 + n_fft // 2)
  >>> dct_filters
  array([[ 0.031,  0.031, ...,  0.031,  0.031],
         [ 0.044,  0.044, ..., -0.044, -0.044],
         ...,
         [ 0.044,  0.044, ..., -0.044, -0.044],
         [ 0.044,  0.044, ...,  0.044,  0.044]])

  >>> import matplotlib.pyplot as plt
  >>> plt.figure()
  >>> librosa.display.specshow(dct_filters, x_axis='linear')
  >>> plt.ylabel('DCT function')
  >>> plt.title('DCT filter bank')
  >>> plt.colorbar()
  >>> plt.tight_layout()
  """
  basis = np.empty((n_filters, n_input))
  basis[0, :] = 1.0 / np.sqrt(n_input)

  samples = np.arange(1, 2 * n_input, 2) * np.pi / (2.0 * n_input)

  for i in range(1, n_filters):
    basis[i, :] = np.cos(i * samples) * np.sqrt(2.0 / n_input)
  return basis

@cache_memory
def mel_filters(sr, n_fft, n_mels=128, fmin=0.0, fmax=None):
  """Create a Filterbank matrix to combine FFT bins into Mel-frequency bins
  Original code: librosa

  Parameters
  ----------
  sr        : number > 0 [scalar]
      sampling rate of the incoming signal

  n_fft     : int > 0 [scalar]
      number of FFT components

  n_mels    : int > 0 [scalar]
      number of Mel bands to generate

  fmin      : float >= 0 [scalar]
      lowest frequency (in Hz)

  fmax      : float >= 0 [scalar]
      highest frequency (in Hz).
      If `None`, use `fmax = sr / 2.0`

  Returns
  -------
  M         : np.ndarray [shape=(n_mels, 1 + n_fft/2)]
      Mel transform matrix

  Examples
  --------
  >>> melfb = mel_filters(22050, 2048)
  >>> melfb
  array([[ 0.   ,  0.016, ...,  0.   ,  0.   ],
         [ 0.   ,  0.   , ...,  0.   ,  0.   ],
         ...,
         [ 0.   ,  0.   , ...,  0.   ,  0.   ],
         [ 0.   ,  0.   , ...,  0.   ,  0.   ]])
  """
  if fmax is None:
    fmax = float(sr) / 2
  # Initialize the weights
  n_mels = int(n_mels)
  weights = np.zeros((n_mels, int(1 + n_fft // 2)))

  # Center freqs of each FFT bin
  fftfreqs = np.linspace(0, float(sr) / 2, int(1 + n_fft // 2),
                         endpoint=True)

  # 'Center freqs' of mel bands - uniformly spaced between limits
  min_mel = hz2mel(fmin)
  max_mel = hz2mel(fmax)
  mel_f = mel2hz(mels=np.linspace(min_mel, max_mel, n_mels + 2))

  fdiff = np.diff(mel_f)
  ramps = np.subtract.outer(mel_f, fftfreqs)

  for i in range(n_mels):
    # lower and upper slopes for all bins
    lower = -ramps[i] / fdiff[i]
    upper = ramps[i + 2] / fdiff[i + 1]

    # .. then intersect them with each other and zero
    weights[i] = np.maximum(0, np.minimum(lower, upper))

  # Slaney-style mel is scaled to be approx constant energy per channel
  enorm = 2.0 / (mel_f[2:n_mels + 2] - mel_f[:n_mels])
  weights *= enorm[:, np.newaxis]

  # Only check weights if f_mel[0] is positive
  if not np.all((mel_f[:-2] == 0) | (weights.max(axis=1) > 0)):
    # This means we have an empty channel somewhere
    print('[WARNING] Empty filters detected in mel frequency basis. '
          'Some channels will produce empty responses. '
          'Try increasing your sampling rate (and fmax) or '
          'reducing n_mels.')
  return weights

@cache_memory
def get_window(window, frame_length, periodic=True):
  ''' Cached version of scipy.signal.get_window '''
  # Funtion
  if hasattr(window, '__call__'):
    return window(frame_length)
  # Window name or scalar
  elif (isinstance(window, (six.string_types, tuple)) or
        np.isscalar(window)):
    return signal.get_window(window, frame_length, fftbins=periodic)
  # Predefined-array
  elif isinstance(window, (np.ndarray, list)):
    if len(window) == frame_length:
      return np.asarray(window)
    raise ValueError('Window size mismatch: '
                     '{:d} != {:d}'.format(len(window), frame_length))
  # Unknown
  else:
    raise ValueError('Invalid window specification: %s' % str(window))

# ===========================================================================
# Array utils
# ===========================================================================
def resample(y, sr_orig, sr_new, axis=0, best_algorithm=True):
  '''
  '''
  sr_orig = int(sr_orig)
  sr_new = int(sr_new)
  if sr_new > sr_orig:
    raise ValueError("Do not support upsampling audio from %d(Hz) to %d(Hz)."
        % (sr_orig, sr_new))
  elif sr_orig != sr_new:
    import resampy
    y = resampy.resample(y, sr_orig=sr_orig, sr_new=sr_new, axis=axis,
            filter='kaiser_best' if best_algorithm else 'kaiser_fast')
  return y

_fnorm1 = lambda x, x_stat, keepdims: x - x_stat.mean(axis=0, keepdims=keepdims)
_fnorm2 = lambda x, x_stat, keepdims: ((x - x_stat.mean(axis=0, keepdims=keepdims)) /
                                       (x_stat.std(axis=0, keepdims=keepdims) + 1e-18))

def mvn(x, varnorm=True, indices=None):
  """ Mean and Variance Normalization
  Normalization is applied on time-axis

  Parameters
  ----------
  x: [t, f]
      [time, frequency]
  varnorm : bool
    if True, normalized by standard deviation
  indices : numpy.ndarray [time,]
    `numpy.bool` array, the speech activities boolean indices,
    which frames will be taken into account for calculating
    the `mean` and `std`

  Note
  ----
  Just standard normalization, not a big deal for its name
  """
  x_stat = x[indices] if indices is not None else x
  if varnorm:
    return _fnorm2(x, x_stat, True)
  else:
    return _fnorm1(x, x_stat, True)

def wmvn(x, w=301, varnorm=True, indices=None):
  """ Windowed - Mean and Variance Normalization
  Normalization is applied on time-axis

  Parameters
  ----------
  x : [t, f]
      [time, frequency]
  w : int
    width of normalization window.
  varnorm : bool
    if True, normalized by standard deviation
  indices : numpy.ndarray [time,]
    `numpy.bool` array, the speech activities boolean indices,
    which frames will be taken into account for calculating
    the `mean` and `std`

  """
  if w < 3 or (w & 1) != 1:
    raise ValueError('Window length should be an odd integer >= 3')
  nobs, ndim = x.shape
  if nobs < w:
    return mvn(x, varnorm=varnorm, indices=indices)
  fnorm = _fnorm2 if varnorm else _fnorm1
  # ====== init ====== #
  hlen = int((w - 1) / 2)
  y = np.zeros((nobs, ndim), dtype=x.dtype)
  # ====== first window ====== #
  # applying indices
  x_stat = x[:w] if indices is None else x[:w][indices[:w]]
  # normalize
  y[:hlen] = fnorm(x[:hlen], x_stat, True)
  # ====== iterate ====== #
  for ix in range(hlen, nobs - hlen):
    # applying indices
    if indices is None:
      x_stat = x[ix - hlen:ix + hlen + 1]
    else:
      sad = indices[ix - hlen:ix + hlen + 1]
      x_stat = x[ix - hlen:ix + hlen + 1][sad]
    # normalize
    y[ix] = fnorm(x[ix], x_stat, False)
  # ====== last window ====== #
  # applying indices
  x_stat = x[nobs - w:] if indices is None else x[nobs - w:][indices[nobs - w:]]
  y[nobs - hlen:nobs] = fnorm(x[nobs - hlen:nobs], x_stat, True)
  return y

def rastafilt(x):
  """ Based on rastafile.m by Dan Ellis
     rows of x = critical bands, cols of x = frame
     same for y but after filtering
     default filter is single pole at 0.94

  The filter is applied on frequency axis

  Parameters
  ----------
  x: [t, f]
      time x frequency
  """
  x = x.T # lazy style to reuse the code from [f, t] libraries
  ndim, nobs = x.shape
  numer = np.arange(-2, 3)
  # careful with division here (float point suggested by Ville Vestman)
  numer = -numer / np.sum(numer * numer)
  denom = [1, -0.94]
  y = np.zeros((ndim, 4))
  z = np.zeros((ndim, 4))
  zi = [0., 0., 0., 0.]
  for ix in range(ndim):
    y[ix, :], z[ix, :] = signal.lfilter(numer, 1, x[ix, :4], zi=zi, axis=-1)
  y = np.zeros((ndim, nobs))
  for ix in range(ndim):
    y[ix, 4:] = signal.lfilter(numer, denom, x[ix, 4:], zi=z[ix, :], axis=-1)[0]
  return y.T

def pre_emphasis(s, coeff=0.97):
  """Pre-emphasis of an audio signal.
  Parameters
  ----------
  s: np.ndarray
      the input vector of signal to pre emphasize
  coeff: float (0, 1)
      coefficience that defines the pre-emphasis filter.
  """
  if s.ndim == 1:
    return np.append(s[0], s[1:] - coeff * s[:-1])
  else:
    return s - np.c_[s[:, :1], s[:, :-1]] * coeff

def smooth(x, win=11, window='hanning'):
  """
  Paramaters
  ----------
  x: 1-D vector
      input signal.
  win: int
      length of window for smoothing, the longer the window, the more details
      are reduced for smoothing.
  window: 'flat', 'hanning', 'hamming', 'bartlett', 'blackman'
      window function, 'flat' for moving average.

  Return
  ------
  y: smoothed vector

  """
  if win < 3:
    return x
  if window not in ['flat', 'hanning', 'hamming', 'bartlett', 'blackman']:
    raise ValueError("Window is on of 'flat', 'hanning', 'hamming', 'bartlett', 'blackman'")
  s = np.concatenate([2 * x[0] - x[win - 1::-1],
                      x,
                      2 * x[-1] - x[-1:-win:-1]], axis=0)
  # moving average
  if window == 'flat':
    w = np.ones(win, dtype='d')
  # windowing
  else:
    w = eval('np.' + window + '(win)')
  y = np.convolve(w / w.sum(), s, mode='same')
  return y[win:-win + 1]

def delta(data, width=9, order=1, axis=0):
  r'''Compute delta features: local estimate of the derivative
  of the input data along the selected axis.
  Original implementation: librosa

  Parameters
  ----------
  data      : np.ndarray
      the input data matrix (e.g. for spectrogram, delta should be applied
      on time-axis).
  width     : int >= 3, odd [scalar]
      Number of frames over which to compute the delta feature
  order     : int > 0 [scalar]
      the order of the difference operator.
      1 for first derivative, 2 for second, etc.
  axis      : int [scalar]
      the axis along which to compute deltas.
      Default is -1 (columns).

  Returns
  -------
  delta_data   : list(np.ndarray) [shape=(d, t) or (d, t + window)]
      delta matrix of `data`.
      return list of deltas

  Examples
  --------
  Compute MFCC deltas, delta-deltas
  >>> mfcc = mfcc(y=y, sr=sr)
  >>> mfcc_delta1, mfcc_delta2 = compute_delta(mfcc, 2)
  '''
  data = np.atleast_1d(data)

  if width < 3 or np.mod(width, 2) != 1:
    raise ValueError('width must be an odd integer >= 3')

  order = int(order)
  if order <= 0:
    raise ValueError('order must be a positive integer')

  half_length = 1 + int(width // 2)
  window = np.arange(half_length - 1., -half_length, -1.)

  # Normalize the window so we're scale-invariant
  window /= np.sum(np.abs(window)**2)

  # Pad out the data by repeating the border values (delta=0)
  padding = [(0, 0)] * data.ndim
  width = int(width)
  padding[axis] = (width, width)
  delta_x = np.pad(data, padding, mode='edge')

  # ====== compute deltas ====== #
  all_deltas = []
  for _ in range(order):
    delta_x = signal.lfilter(window, 1, delta_x, axis=axis)
    all_deltas.append(delta_x)
  # ====== Cut back to the original shape of the input data ====== #
  trim_deltas = []
  for delta_x in all_deltas:
    idx = [slice(None)] * delta_x.ndim
    idx[axis] = slice(- half_length - data.shape[axis], - half_length)
    delta_x = delta_x[idx]
    trim_deltas.append(delta_x.astype('float32'))
  return trim_deltas[0] if order == 1 else trim_deltas

def shifted_deltas(x, N=7, d=1, P=3, k=7):
  """ Calculate Shifted Delta Coefficients
  (suggested applying on time-dimension) for MFCCs

  Parameters
  ----------
  x: [t, f]
      time x frequency
  """
  x = x.T
  if d < 1:
    raise ValueError('d should be an integer >= 1')
  nobs = x.shape[1]
  x = x[:N]
  w = 2 * d + 1
  dx = delta(x, w, order=1, axis=-1)
  sdc = np.empty((k * N, nobs))
  sdc[:] = np.tile(dx[:, -1], k).reshape(k * N, 1)
  for ix in range(k):
    if ix * P > nobs:
      break
    sdc[ix * N:(ix + 1) * N, :nobs - ix * P] = dx[:, ix * P:nobs]
  return sdc.T

@cache_memory('__strict__')
def pad_center(data, size, axis=-1, **kwargs):
  '''Wrapper for numpy.pad to automatically center an array prior to padding.
  This is analogous to `str.center()`

  Parameters
  ----------
  data : numpy.ndarray
      Vector to be padded and centered

  size : int >= len(data) [scalar]
      Length to pad `data`

  axis : int
      Axis along which to pad and center the data

  kwargs : additional keyword arguments
    arguments passed to `numpy.pad()`

  Returns
  -------
  data_padded : numpy.ndarray
      `data` centered and padded to length `size` along the
      specified axis

  Raises
  ------
  ParameterError
      If `size < data.shape[axis]`

  See Also
  --------
  numpy.pad
  '''
  kwargs.setdefault('mode', 'constant')

  n = data.shape[axis]

  lpad = int((size - n) // 2)

  lengths = [(0, 0)] * data.ndim
  lengths[axis] = (lpad, int(size - n - lpad))

  if lpad < 0:
    raise ValueError(('Target size ({:d}) must be '
                      'at least input size ({:d})').format(size, n))
  return np.pad(data, lengths, **kwargs)

def one_hot(y, nb_classes=None, dtype='float32'):
  '''Convert class vector (integers from 0 to nb_classes)
  to binary class matrix, for use with categorical_crossentropy

  Note
  ----
  if any class index in y is smaller than 0, then all of its one-hot
  values is 0.
  '''
  if 'int' not in str(y.dtype):
    y = y.astype('int32')
  if nb_classes is None:
    nb_classes = np.max(y) + 1
  else:
    nb_classes = int(nb_classes)
  return np.eye(nb_classes, dtype=dtype)[y]

def pad_sequences(sequences, maxlen=None, dtype='int32',
                  padding='pre', truncating='pre', value=0.,
                  transformer=None):
  """Pads each sequence to the same length:
  the length of the longest sequence.

  If maxlen is provided, any sequence longer
  than maxlen is truncated to maxlen.
  Truncation happens off either the beginning (default) or
  the end of the sequence.

  Supports post-padding and pre-padding (default).

  Parameters
  ----------
  sequences: list
      a list that contains a list of object
  maxlen: int
      maximum length of each individual sequence
  dtype: np.dtype
      desire data type of output array
  padding: 'pre' or 'post'
      pad either before or after each sequence.
  truncating: 'pre' or 'post'
      remove values from sequences larger than maxlen either
      in the beginning or in the end of the sequence
  value: object
      padding object
  transformer: call-able
      a function transform each element in sequence into desire value
      (e.g. a dictionary)

  Returns
  -------
  numpy array with dimensions (number_of_sequences, maxlen)
  """
  # ====== check valid input ====== #
  if truncating not in ('pre', 'post'):
    raise ValueError('truncating must be "pre" or "post", given value is %s'
                     % truncating)
  if padding not in ('pre', 'post'):
    raise ValueError('padding must be "pre" or "post", given value is %s'
                     % padding)
  if transformer is None:
    transformer = lambda x: x
  if not hasattr(transformer, '__call__'):
    raise ValueError('transformer must be call-able, but given value is %s' %
                     type(transformer))
  # ====== processing ====== #
  if maxlen is None:
    maxlen = int(max(len(s) for s in sequences))
  nb_samples = len(sequences)
  value = np.cast[dtype](value)
  X = np.full(shape=(nb_samples, maxlen), fill_value=value, dtype=dtype)
  for idx, s in enumerate(sequences):
    s = [transformer(_) for _ in s]
    if len(s) == 0: continue # empty list
    # check truncating
    if len(s) >= maxlen:
      slice_ = slice(None, None)
      s = s[-maxlen:] if truncating == 'pre' else s[:maxlen]
    # check padding
    elif len(s) < maxlen:
      slice_ = slice(-len(s), None) if padding == 'pre' else slice(None, len(s))
    # assign value
    X[idx, slice_] = np.asarray(s, dtype=dtype)
  return X

def stack_frames(X, frame_length, step_length=None,
                 keep_length=False, make_contigous=False):
  """ Stack consecutive frames into single vector, each operation
  is shifted by `step_length`

  Parameters
  ----------
  X: numpy.ndarray
      2D arrray
  frame_length: int
      number of frames will be stacked into 1 sample.
  step_length: {int, None} (default: None)
      number of shifted frame after each stacking operation,
      if None, its value equals to `frame_length // 2`
  keep_length: bool
      if True, padding zeros to begin and end of `X` to
      make the output array has the same length as original
      array.
  make_contigous: bool
      if True, use `numpy.ascontiguousarray` to ensure input `X`
      is contiguous.

  Example
  -------
  >>> X = [[ 0  1]
  ...      [ 2  3]
  ...      [ 4  5]
  ...      [ 6  7]
  ...      [ 8  9]
  ...      [10 11]
  ...      [12 13]
  ...      [14 15]
  ...      [16 17]
  ...      [18 19]]
  >>> frame_length = 5
  >>> step_length = 2
  >>> stack_frames(X, frame_length, step_length)
  >>> [[ 0  1  2  3  4  5  6  7  8  9]
  ...  [ 4  5  6  7  8  9 10 11 12 13]
  ...  [ 8  9 10 11 12 13 14 15 16 17]]
  """
  if frame_length > len(X) and keep_length is False:
    raise ValueError("`frame_length=%d` is greater than the length of input matrix %s;"
                     "`keep_length` must be set to True to allow padding." %
                     (frame_length, str(X.shape)))
  if keep_length:
    if step_length != 1:
      raise ValueError("`keep_length` is only supported when `step_length` = 1.")
    add_frames = (int(np.ceil(frame_length / 2)) - 1) * 2 + \
        (1 if frame_length % 2 == 0 else 0)
    right = add_frames // 2
    left = add_frames - right
    X = np.pad(X,
               pad_width=((left, right),) + ((0, 0),) * (X.ndim - 1),
               mode='constant')
  # ====== check input ====== #
  assert X.ndim == 2, "Only support 2D matrix for stacking frames."
  if not X.flags['C_CONTIGUOUS']:
    if make_contigous:
      X = np.ascontiguousarray(X)
    else:
      raise ValueError('Input buffer must be contiguous.')
  # ====== stacking ====== #
  frame_length = int(frame_length)
  if step_length is None:
    step_length = frame_length // 2
  shape = (1 + (X.shape[0] - frame_length) // step_length,
           frame_length * X.shape[1])
  strides = (X.strides[0] * step_length, X.strides[1])
  return np.lib.stride_tricks.as_strided(X, shape=shape, strides=strides)

def segment_axis(a, frame_length=2048, step_length=512, axis=0,
                 end='cut', pad_value=0, pad_mode='post'):
  """Generate a new array that chops the given array along the given axis
  into overlapping frames.

  This method has been implemented by Anne Archibald,
  as part of the talk box toolkit
  example::

      segment_axis(arange(10), 4, 2)
      array([[0, 1, 2, 3],
         ( [2, 3, 4, 5],
           [4, 5, 6, 7],
           [6, 7, 8, 9]])

  Parameters
  ----------
  a: numpy.ndarray
      the array to segment
  frame_length: int
      the length of each frame
  step_length: int
      the number of array elements by which the frames should overlap
  axis: int, None
      the axis to operate on; if None, act on the flattened array
  end: 'cut', 'wrap', 'pad'
      what to do with the last frame, if the array is not evenly
          divisible into pieces. Options are:
          - 'cut'   Simply discard the extra values
          - 'wrap'  Copy values from the beginning of the array
          - 'pad'   Pad with a constant value
  pad_value: int
      the value to use for end='pad'
  pad_mode: 'pre', 'post'
      if "pre", padding or wrapping at the beginning of the array.
      if "post", padding or wrapping at the ending of the array.

  Return
  ------
  a ndarray

  The array is not copied unless necessary (either because it is unevenly
  strided and being flattened or because end is set to 'pad' or 'wrap').

  Note
  ----
  Modified work and error fixing Copyright (c) TrungNT

  """
  if axis is None:
    a = np.ravel(a) # may copy
    axis = 0

  length = a.shape[axis]
  overlap = frame_length - step_length

  if overlap >= frame_length:
    raise ValueError("frames cannot overlap by more than 100%")
  if overlap < 0 or frame_length <= 0:
    raise ValueError("overlap must be nonnegative and length must" +
                     "be positive")

  if length < frame_length or (length - frame_length) % (frame_length - overlap):
    if length > frame_length:
      roundup = frame_length + (1 + (length - frame_length) // (frame_length - overlap)) * (frame_length - overlap)
      rounddown = frame_length + ((length - frame_length) // (frame_length - overlap)) * (frame_length - overlap)
    else:
      roundup = frame_length
      rounddown = 0
    assert rounddown < length < roundup
    assert roundup == rounddown + (frame_length - overlap) \
    or (roundup == frame_length and rounddown == 0)
    a = a.swapaxes(-1, axis)

    if end == 'cut':
      a = a[..., :rounddown]
    elif end in ['pad', 'wrap']: # copying will be necessary
      s = list(a.shape)
      s[-1] = roundup
      b = np.empty(s, dtype=a.dtype)
      # pre-padding
      if pad_mode == 'post':
        b[..., :length] = a
        if end == 'pad':
          b[..., length:] = pad_value
        elif end == 'wrap':
          b[..., length:] = a[..., :roundup - length]
      # post-padding
      elif pad_mode == 'pre':
        b[..., -length:] = a
        if end == 'pad':
          b[..., :(roundup - length)] = pad_value
        elif end == 'wrap':
          b[..., :(roundup - length)] = a[..., :roundup - length]
      # error
      else:
        raise RuntimeError("No support for pad mode: %s" % pad_mode)
      a = b
    a = a.swapaxes(-1, axis)
    length = a.shape[0] # update length

  if length == 0:
    raise ValueError("Not enough data points to segment array " +
            "in 'cut' mode; try 'pad' or 'wrap'")
  assert length >= frame_length
  assert (length - frame_length) % (frame_length - overlap) == 0
  n = 1 + (length - frame_length) // (frame_length - overlap)
  s = a.strides[axis]
  newshape = a.shape[:axis] + (n, frame_length) + a.shape[axis + 1:]
  newstrides = a.strides[:axis] + ((frame_length - overlap) * s, s) + a.strides[axis + 1:]

  try:
    return np.ndarray.__new__(np.ndarray, strides=newstrides,
                              shape=newshape, buffer=a, dtype=a.dtype)
  except TypeError:
    a = a.copy()
    # Shape doesn't change but strides does
    newstrides = a.strides[:axis] + ((frame_length - overlap) * s, s) \
    + a.strides[axis + 1:]
    return np.ndarray.__new__(np.ndarray, strides=newstrides,
                              shape=newshape, buffer=a, dtype=a.dtype)

# ===========================================================================
# Fourier transform
# ===========================================================================
def get_energy(frames, log=True):
  """ Calculate frame-wise energy

  Parameters
  ----------
  frames: ndarray
      framed signal with shape (nb_frames x window_length)
  log: bool
      if True, return log energy of each frames

  Return
  ------
  E : ndarray [shape=(nb_frames,), dtype=float32]
  """
  log_energy = (frames**2).sum(axis=1)
  log_energy = np.where(log_energy == 0., np.finfo(np.float32).eps,
                        log_energy)
  if log:
    log_energy = np.log(log_energy)
  return np.expand_dims(log_energy.astype('float32'), -1)

def stft(y,
         frame_length=None, step_length=None, n_fft=None,
         window='hann', scale=None,
         padding=False, energy=False):
  """Short-time Fourier transform (STFT)

  Returns a complex-valued matrix D such that
      `np.abs(D[f, t])` is the magnitude of frequency bin `f`
      at frame `t`

      `np.angle(D[f, t])` is the phase of frequency bin `f`
      at frame `t`

  Parameters
  ----------
  y : numpy.ndarray [shape=(n_samples,)] or [shape=(n_frames, frame_length)]
      the input signal (audio time series) or framed signal in case
      of 2-D matrix with `shape[1] > 2`

  frame_length: int
      number of samples point for 1 frame

  step_length: int
      number of samples point for 1 step (when shifting the frames)
      If unspecified, defaults `frame_length / 4`.

  n_fft: int > 0 [scalar]
      FFT window size
      If not provided, uses the smallest power of 2 enclosing `frame_length`.

  window : string, tuple, number, function, or np.ndarray [shape=(n_fft,)]
      - a window specification (string, tuple, or number);
        see `scipy.signal.get_window`
      - a window function, such as `scipy.signal.hanning`
      - a vector or array of length `n_fft`

  scale : {None, float}
      re-scale the STFT matrix after windowing, it is important factor
      for reconstruct original signal using iSTFT

  padding: boolean
      - If `True`, the signal `y` is padded so that frame
        `D[:, t]` is centered at `y[t * step_length]`.
      - If `False`, then `D[:, t]` begins at `y[t * step_length]`

  energy: bool
      if True, return log-frame-wise energy

  Returns
  -------
  D : np.ndarray [shape=(t, 1 + n_fft/2), dtype=complex64]
      STFT matrix
  log_energy : ndarray [shape=(t,), dtype=float32]
      (log) energy of each frame

  Note
  ----
  This implementation have been tested to achieve slightly better speed
  than scipy implementation
  """
  # ====== check input signal ====== #
  if y.ndim == 2 and y.shape[1] > 2:
    y, frames = None, y
  else:
    y, frames = y, None
  # ====== check others ====== #
  # check frame_length
  if frame_length is None:
    if frames is not None:
      frame_length = frames.shape[1]
    else:
      raise ValueError("When `frame_length` is None, `frames` must be provided")
  else:
    frame_length = int(frame_length)
  # check step_length
  if step_length is None:
    step_length = frame_length // 4
  else:
    step_length = int(step_length)
  # check n_fft
  if n_fft is None:
    n_fft = int(2**np.ceil(np.log(frame_length) / np.log(2.0)))
  elif n_fft < frame_length:
    raise ValueError('n_fft must be greater than or equal to `frame_length`.')
  # ====== doing framing on raw signal ====== #
  if frames is None:
    # check if padding zeros
    if padding:
      y = np.pad(y, int(frame_length // 2), mode='constant')
    # framing the signal
    shape = y.shape[:-1] + (y.shape[-1] - frame_length + 1, frame_length)
    strides = y.strides + (y.strides[-1],)
    y_frames = np.lib.stride_tricks.as_strided(y, shape=shape, strides=strides)
    if y_frames.ndim > 2:
      y_frames = np.rollaxis(y_frames, 1)
    # [n_frames, frame_length]
    y_frames = y_frames[::step_length]
  else:
    y_frames = frames
  # ====== prepare the window function ====== #
  if window is not None:
    fft_window = get_window(window, frame_length, periodic=True
        ).reshape(1, -1)
    y_frames = fft_window * y_frames
    # scaling factor
    scale = np.sqrt(1.0 / fft_window.sum()**2) if scale is None else float(scale)
  else:
    scale = np.sqrt(1.0 / frame_length**2) if scale is None else float(scale)
  # ====== calculate frames energy ====== #
  if energy:
    log_energy = get_energy(y_frames, log=True).astype('float32')
  # ====== STFT matrix ====== #
  # norm='ortho' ?
  S = np.fft.rfft(a=y_frames, n=n_fft, axis=-1)
  # this scale is important for iSTFT reconstruct original signal
  if scale is not None:
    S *= scale
  # return in form (t, d)
  if energy:
    return S, log_energy
  return S


def istft(stft_matrix, frame_length, step_length=None,
          window='hann', padding=False):
  """
  Inverse short-time Fourier transform (ISTFT).

  Converts a complex-valued spectrogram `stft_matrix` to time-series `y`
  by minimizing the mean squared error between `stft_matrix` and STFT of
  `y` as described in [1]_.

  In general, window function, hop length and other parameters should be same
  as in stft, which mostly leads to perfect reconstruction of a signal from
  unmodified `stft_matrix`.

  .. [1] D. W. Griffin and J. S. Lim,
      "Signal estimation from modified short-time Fourier transform,"
      IEEE Trans. ASSP, vol.32, no.2, pp.236–243, Apr. 1984.

  Parameters
  ----------
  stft_matrix : np.ndarray [shape=(1 + n_fft/2, t)]
      STFT matrix from `stft`
  frame_length: int
      number of samples point for 1 frame
  step_length: int
      number of samples point for 1 step (when shifting the frames)
      If unspecified, defaults `frame_length / 4`.
  window      : string, tuple, number, function, np.ndarray [shape=(n_fft,)]
      - a window specification (string, tuple, or number);
        see `scipy.signal.get_window`
      - a window function, such as `scipy.signal.hanning`
      - a user-specified window vector of length `n_fft`
  padding: boolean
      - If `True`, the signal `y` is padded so that frame
        `D[:, t]` is centered at `y[t * step_length]`.
      - If `False`, then `D[:, t]` begins at `y[t * step_length]`

  Returns
  -------
  y : np.ndarray [shape=(n,), dtype=float32]
      time domain signal reconstructed from `stft_matrix`
  """
  # ====== check arguments ====== #
  frame_length = int(frame_length)
  if step_length is None:
    step_length = frame_length // 4
  else:
    step_length = int(step_length)
  n_fft = 2 * (stft_matrix.shape[1] - 1)
  # ====== use scipy here ====== #
  try:
    from scipy.signal.spectral import istft as _istft
  except ImportError:
    raise RuntimeError("`istft` requires scipy version >= 0.19")
  return _istft(stft_matrix, fs=1.0, window=window,
                nperseg=frame_length, noverlap=frame_length - step_length, nfft=n_fft,
                input_onesided=True, boundary=padding,
                time_axis=0, freq_axis=-1)[-1]

def power_spectrogram(S, power=2.0):
  """ Extracting power spectrum from a complex-STFT array

  Parameters
  ----------
  S : array [nb_samples, n_fft]
    complex type or real type
  power : float
    factor for converting spectrogram to power spectrum

  Note
  ----
  To convert power spectrum to Decibel (db)
  >>> signal.power2db(power_spectrogram(S, power=2.0), top_db=80.0)

  """
  power = int(power)
  # ====== extract the magnitude spectrogram ====== #
  if 'complex' in str(S.dtype): # get magnitude from STFT
    spec = np.abs(S)
  else:
    spec = S
  # ====== power ====== #
  if power > 1:
    spec = np.power(spec, power)
  return spec

def mels_spectrogram(spec, sr, n_mels,
                     fmin=64, fmax=None, top_db=80.0):
  """ Extracting mel-filter bands from power spectrum
  (i.e. the output from function
  `odin.preprocessing.signal.power_spectrogram`)

  Parameters
  ----------
  spec : array [nb_samples, n_fft]
    power spectrum array
  sr : int
    sample rate
  fmin: int
      min frequency for mel-filter bands
  fmax: int, or None
      maximum frequency for mel-filter bands.
      If None, usng `sr / 2` as fmax
  top_db: int
      maximum deciben

  """
  # ====== check arguments ====== #
  n_fft = int(2 * (spec.shape[1] - 1))
  # check fmax
  if sr is None and fmax is None:
    fmax = 4000
  else:
    fmax = sr // 2 if fmax is None else int(fmax)
  # check fmin
  fmin = int(fmin)
  if fmin >= fmax:
    raise ValueError("fmin must < fmax, but fmin=%d and fmax=%d" %
                     (fmin, fmax))
  # ====== mel transform ====== #
  mel_basis = mel_filters(sr,
      n_fft=n_fft, n_mels=24 if n_mels is None else int(n_mels),
      fmin=fmin, fmax=fmax)
  # transpose to (nb_samples; nb_mels)
  mel_spec = np.dot(mel_basis, spec.T)
  mel_spec = mel_spec.T
  mel_spec = power2db(mel_spec, top_db=top_db)
  return mel_spec

def ceps_spectrogram(mspec, n_ceps, remove_first_coef=True):
  """ Compute the MFCCs coefficients (cepstrum analysis)
  from extracted mel-filter bands spectrogram
  (i.e. output from `odin.preprocessing.signal.mels_spectrogram`)

  Parameters
  ----------
  mspec : array [nb_samples, n_fft]
    mels-spectrogram array
  n_ceps : int
    number of ceptrum for cepstral analysis
  remove_first_coef : bool
    if True remove the first coefficient of the extracted MFCCs

  """
  if remove_first_coef:
    n_ceps = int(n_ceps) + 1
    dct_basis = dct_filters(n_ceps, mspec.shape[1])
    mfcc = np.dot(dct_basis, mspec.T)[1:, :].T
  else:
    n_ceps = int(n_ceps)
    dct_basis = dct_filters(n_ceps, mspec.shape[1])
    mfcc = np.dot(dct_basis, mspec.T).T
  return mfcc

def spectra(sr, frame_length, y=None, S=None,
            step_length=None, n_fft=512, window='hann',
            n_mels=None, n_ceps=None,
            fmin=64, fmax=None,
            top_db=80.0, power=2.0, log=True,
            padding=False):
  """Compute spectra information from STFT matrix or a power spectrogram,
  The extracted spectra include:
  * log-power spectrogram
  * mel-scaled spectrogram.
  * MFCC (cepstrum analysis)

  If a spectrogram input `S` is provided, then it is mapped directly onto
  the mel basis `mel_f` by `mel_f.dot(S)`.

  If a time-series input `y, sr` is provided, then its magnitude spectrogram
  `S` is first computed, and then mapped onto the mel scale by
  `mel_f.dot(S**power)`.  By default, `power=2` operates on a power spectrum.

  Parameters
  ----------
  sr : number > 0 [scalar]
      sampling rate of `y`
  frame_length: int
      number of samples point for 1 frame
  y : np.ndarray [shape=(n,)] or None
      audio time-series
  S : np.ndarray [shape=(d, t)]
      spectrogram or complex STFT
  step_length: int
      number of samples point for 1 step (when shifting the frames)
      If unspecified, defaults `frame_length / 4`.
  n_fft : int > 0 [scalar]
      length of the FFT window
  window      : string, tuple, number, function, np.ndarray [shape=(n_fft,)]
      - a window specification (string, tuple, or number);
        see `scipy.signal.get_window`
      - a window function, such as `scipy.signal.hanning`
      - a user-specified window vector of length `n_fft`
  n_mels: int, or None
      number of mel-filter bands
  n_ceps: int, or None
      number of ceptrum for cepstral analysis
  fmin: int
      min frequency for mel-filter bands
  fmax: int, or None
      maximum frequency for mel-filter bands.
      If None, usng `sr / 2` as fmax
  top_db: int
      maximum deciben
  power : float > 0 [scalar]
      Exponent for the magnitude spectrogram.
      e.g., 1 for energy (or magnitude), 2 for power, etc.
  log: bool
      if True, convert all power spectrogram to DB
  padding : bool
      - If `True`, the signal `y` is padded so that frame
        `D[:, t]` is centered at `y[t * step_length]`.
      - If `False`, then `D[:, t]` begins at `y[t * step_length]`

  Returns
  -------
  S : np.ndarray [shape=(t, nb_melfilters)]
      Mel spectrogram

  Note
  ----
  `log` and `power` don't work for `sptk` backend
  This method is optimized for speed and memory when extracting
  a sequence of speech features at once.

  """
  mel_spec = None
  mfcc = None
  log_energy = None
  # ====== compute STFT if needed ====== #
  if S is None:
    S, log_energy = stft(y, frame_length=frame_length,
                         step_length=step_length, n_fft=n_fft,
                         window=window, padding=padding, energy=True)
  n_fft = int(2 * (S.shape[1] - 1))
  # ====== check arguments ====== #
  power = int(power)
  # check fmax
  if sr is None and fmax is None:
    fmax = 4000
  else:
    fmax = sr // 2 if fmax is None else int(fmax)
  # check fmin
  fmin = int(fmin)
  if fmin >= fmax:
    raise ValueError("fmin must < fmax, but fmin=%d and fmax=%d" %
                     (fmin, fmax))
  # ====== extract the basic spectrogram ====== #
  if 'complex' in str(S.dtype): # get magnitude from STFT
    spec = np.abs(S)
  if power > 1:
    spec = np.power(spec, power)
  # ====== extrct mel-filter-bands features ====== #
  if n_mels is not None or n_ceps is not None:
    mel_spec = mels_spectrogram(spec, sr, n_mels)
  # ====== extract cepstrum features ====== #
  # extract MFCC
  if n_ceps is not None:
    mfcc = ceps_spectrogram(mel_spec, n_ceps)
  # applying log to convert to db
  if log:
    spec = power2db(spec, top_db=top_db)
  # ====== return result ====== #
  results = {}
  results['spec'] = spec.astype('float32')
  results['energy'] = log_energy
  results['mspec'] = None if mel_spec is None else mel_spec.astype('float32')
  results['mfcc'] = None if mfcc is None else mfcc.astype('float32')
  return results


# ===========================================================================
# invert spectrogram
# ===========================================================================
def ispec(spec, frame_length, step_length=None, window="hann",
          nb_iter=48, normalize=True,
          db=False, padding=False,
          de_preemphasis=0.97):
  """ Using Griffin-Lim algorithm for Inverting
  (power) spectrogram back to raw waveform

  Parameters
  ----------
  spec : np.ndarray [shape=(t, n_fft / 2 + 1)]
      magnitude, power, or DB spectrogram of STFT
  frame_length: int
      number of samples point for 1 frame
  step_length: int
      number of samples point for 1 step (when shifting the frames)
      If unspecified, defaults `frame_length / 4`.
  window : string, tuple, number, function, or np.ndarray [shape=(n_fft,)]
      - a window specification (string, tuple, or number);
        see `scipy.signal.get_window`
      - a window function, such as `scipy.signal.hanning`
      - a vector or array of length `n_fft`
  nb_iter: int
      number of iteration, the higher the better audio quality
  db: bool
      if the given spectrogram is in decibel (dB) units (used logarithm)
  normalize: bool
      normalize output raw signal to have mean=0., and std=1.

  """
  spec = spec.astype('float64')
  # ====== check arguments ====== #
  frame_length = int(frame_length)
  if step_length is None:
    step_length = frame_length // 4
  else:
    step_length = int(step_length)
  if frame_length < step_length:
    raise ValueError('frame_length=%d < step_length=%d' %
      (frame_length, step_length))
  n_fft = (spec.shape[1] - 1) * 2
  # ====== convert to power spectrogram ====== #
  if db:
    spec = db2power(spec)
  # ====== iterative estmate best phase ====== #
  X_best = copy.deepcopy(spec)
  for i in range(nb_iter):
    X_t = istft(X_best, frame_length=frame_length, step_length=step_length,
                window=window, padding=padding)
    est = stft(X_t, frame_length=frame_length, step_length=step_length,
               n_fft=n_fft, window=window, padding=padding, energy=False)
    phase = est / np.maximum(1e-8, np.abs(est))
    X_best = spec * phase
  X_t = istft(X_best, frame_length=frame_length, step_length=step_length,
              window=window, padding=padding)
  # ====== de-preemphasis ====== #
  if isinstance(de_preemphasis, Number) and 0. < de_preemphasis < 1.:
    X_t = signal.lfilter([1], [1, -de_preemphasis], X_t)
  y = np.real(X_t)
  if normalize:
    y = y[1000:-1000]
    y = (y - y.mean()) / y.std()
  return y

# ===========================================================================
# F0 analysis
# ===========================================================================
def pitch_track(y, sr, step_length, fmin=60.0, fmax=260.0, threshold=0.3,
                otype="pitch", algorithm='swipe'):
  """

  Parameters
  ----------
  y : array
      A whole audio signal
  sr : int
      Sampling frequency.
  step_length : int
      Hop length.
  fmin : float, optional
      Minimum fundamental frequency. Default is 60.0
  fmax : float, optional
      Maximum fundamental frequency. Default is 260.0
  threshold : float, optional
      Voice/unvoiced threshold. Default is 0.3 (as suggested for SWIPE)
      Threshold >= 1.0 is suggested for RAPT
  otype : str or int, optional
      Output format
          (0) pitch
          (1) f0
          (2) log(f0)
      Default is f0.
  algorithm: 'swipe', 'rapt', 'avg'
      SWIPE - A Saw-tooth Waveform Inspired Pitch Estimation.
      RAPT - a robust algorithm for pitch tracking.
      avg - apply swipe and rapt at the same time, then take average.
      Default is 'SWIPE'

  Note
  ----
  by default, the pitch tracking algorithm always center the signal before
  processing
  """
  try:
    import pysptk
  except ImportError:
    raise RuntimeError("Pitch tracking requires pysptk library.")
  # ====== check arguments ====== #
  algorithm = str(algorithm).lower()
  if algorithm not in ('rapt', 'swipe', 'avg'):
    raise ValueError("'algorithm' argument must be: 'rapt', 'swipe' or 'avg'")
  otype = str(otype)
  if otype not in ('f0', 'pitch'):
    raise ValueError("Support 'otype' include: 'f0', 'pitch'")
  # ====== Do it ====== #
  sr = int(sr)
  if algorithm == 'avg':
    y1 = pysptk.swipe(y.astype(np.float64), fs=sr, hopsize=step_length,
                      threshold=threshold, min=fmin, max=fmax, otype=otype)
    y2 = pysptk.rapt(y.astype(np.float32), fs=sr, hopsize=step_length,
                     voice_bias=threshold, min=fmin, max=fmax, otype=otype)
    y = (y1 + y2) / 2
  elif algorithm == 'swipe':
    y = pysptk.swipe(y.astype(np.float64), fs=sr, hopsize=step_length,
                     threshold=threshold, min=fmin, max=fmax, otype=otype)
  else:
    y = pysptk.rapt(y.astype(np.float32), fs=sr, hopsize=step_length,
                    voice_bias=threshold, min=fmin, max=fmax, otype=otype)
  return y.astype('float32')



import numpy as np
import tensorflow as tf

from odin.autoconfig import EPS
from odin.utils import is_number
from odin.backend.role import (return_roles, DifferentialLoss, add_roles)
from odin.backend.tensor import is_tensor, to_nonzeros, dimshuffle

# ===========================================================================
# Similarity measurement
# ===========================================================================
@return_roles(roles=DifferentialLoss)
def contrastive_loss(y_true, y_pred, margin=1, name=None):
  """
  Reference
  ---------
  [1] http://yann.lecun.com/exdb/publis/pdf/hadsell-chopra-lecun-06.pdf
  """
  # ====== tensorflow Tensor ====== #
  if is_tensor(y_true) and is_tensor(y_pred):
    with tf.name_scope(name, 'contrastive_loss', [y_true, y_pred]):
      loss = tf.reduce_mean(
          y_true * tf.square(y_pred) +
          (1 - y_true) * tf.square(tf.maximum(margin - y_pred, 0)))
  # ====== ndarray or similar ====== #
  else:
    loss = np.mean(y_true * np.square(y_pred) +
                  (1 - y_true) * np.square(np.maximum(margin - y_pred, 0)))
  return loss

@return_roles(roles=DifferentialLoss)
def triplet_loss(y_true, y_pred):
  pass

def contrastive_loss_andre(left_feature, right_feature, label, margin):
  """
  Compute the contrastive loss as in
  https://gitlab.idiap.ch/biometric/xfacereclib.cnn/blob/master/xfacereclib/cnn/scripts/experiment.py#L156
  With Y = [-1 +1] --> [POSITIVE_PAIR NEGATIVE_PAIR]
  L = log( m + exp( Y * d^2)) / N
  **Parameters**
   left_feature: First element of the pair
   right_feature: Second element of the pair
   label: Label of the pair (0 or 1)
   margin: Contrastive margin
  **Returns**
   Return the loss operation
  """

  with tf.name_scope("contrastive_loss_andre"):
    label = tf.to_float(label)
    d = compute_euclidean_distance(left_feature, right_feature)

    loss = tf.log(tf.exp(tf.mul(label, d)))
    loss = tf.reduce_mean(loss)

    # Within class part
    genuine_factor = tf.mul(label - 1, 0.5)
    within_class = tf.reduce_mean(tf.log(tf.exp(tf.mul(genuine_factor, d))))

    # Between class part
    impostor_factor = tf.mul(label + 1, 0.5)
    between_class = tf.reduce_mean(tf.log(tf.exp(tf.mul(impostor_factor, d))))

    # first_part = tf.mul(one - label, tf.square(d))  # (Y-1)*(d^2)
    return loss, between_class, within_class

@return_roles(DifferentialLoss)
def cosine_similarity(y_true, y_pred, weights=1.0,
                      unit_norm=True, one_vs_all=True,
                      name=None):
  """
  Parameters
  ----------
  y_true : {Tensor, ndarray}
      enrollment vectors, one sample per row
      (i.e. shape=(nb_samples, nb_features))
  y_pred : {Tensor, ndarray}
      test vectors, one sample per row
      (i.e. shape=(nb_samples, nb_features))
  unit_norm : bool (default: True)
      if True, normalize length of each vector to 1
  one_vs_all : bool (default: True)
      if True, calculate the similarity of one to all other
      samples, otherwise, it is `one_vs_one`

  Return
  ------
  scores : score matrix,
      comparing all models against all tests

  """
  # ====== tensorflow Tensor ====== #
  if is_tensor(y_true) and is_tensor(y_pred):
    with tf.name_scope(name, "cosine_similarity", (y_true, y_pred, weights)):
      y_pred.shape.assert_is_compatible_with(y_true.shape)
      if unit_norm:
        y_true /= to_nonzeros(tf.linalg.norm(y_true, ord=2, axis=-1, keep_dims=True), 1.)
        y_pred /= to_nonzeros(tf.linalg.norm(y_pred, ord=2, axis=-1, keep_dims=True), 1.)
      if one_vs_all:
        scores = tf.matmul(tf.transpose(y_true), y_pred)
      else:
        scores = 1 - tf.reduce_sum(tf.multiply(y_true, y_pred),
                                   axis=(-1,), keep_dims=True)
  # ====== ndarray or similar ====== #
  else:
    assert y_true.shape == y_pred.shape
    if unit_norm:
      y_true /= to_nonzeros(np.linalg.norm(y_true, ord=2, axis=-1, keepdims=True), 1.)
      y_pred /= to_nonzeros(np.linalg.norm(y_pred, ord=2, axis=-1, keepdims=True), 1.)
    if one_vs_all:
      scores = np.dot(y_true.T, y_pred)
    else:
      scores = np.sum(y_true * y_pred,
                      axis=(-1,), keepdims=True)
  return scores

# ===========================================================================
# Cross-entropy variations
# ===========================================================================
@return_roles(DifferentialLoss)
def bayes_crossentropy(y_true, y_pred, nb_classes=None, reduction=tf.reduce_mean,
                       name=None):
  with tf.name_scope(name, "bayes_crossentropy", [y_true, y_pred]):
    y_pred_shape = y_pred.shape
    if y_pred_shape.ndims == 1 or y_pred_shape[-1].value == 1:
      if y_pred_shape.ndims == 1:
        y_pred = tf.expand_dims(y_pred, -1)
      y_pred0 = 1. - y_pred
      y_pred = tf.concat([y_pred0, y_pred], axis=-1)
    # get number of classes
    if y_true.shape.ndims == 1:
      if nb_classes is None:
        raise Exception('y_pred and y_true must be one_hot encoded, '
                        'otherwise you have to provide nb_classes.')
      y_true = tf.one_hot(y_true, depth=nb_classes)
    elif nb_classes is None:
      nb_classes = y_true.shape[1].value
    # avoid numerical instability with _EPSILON clipping
    y_pred = tf.clip_by_value(y_pred, EPS, 1.0 - EPS)
    # ====== check distribution ====== #
    distribution = tf.reduce_sum(y_true, axis=0)
    # probability distribution of each class
    prob_distribution = dimshuffle(distribution / tf.reduce_sum(distribution),
                                   ('x', 0))
    # we need to clip the prior probability distribution also
    prob_distribution = tf.clip_by_value(prob_distribution, EPS, 1.0 - EPS)
    # ====== init confusion info loss ====== #
    # weighted by y_true
    loss = y_true * tf.log(y_pred)
    loss = - 1 / nb_classes * tf.reduce_sum(loss / prob_distribution, axis=1)
    return reduction(loss)

@return_roles(DifferentialLoss)
def bayes_binary_crossentropy(y_true, y_pred):
  y_pred = tf.concat([1 - y_pred, y_pred], axis=-1)
  y_true = tf.one_hot(tf.cast(y_true, 'int32'), depth=2)
  return bayes_crossentropy(y_pred, y_true, nb_classes=2)


# ===========================================================================
# Variational
# ===========================================================================
def jacobian_regularize(hidden, params):
  """ Computes the jacobian of the hidden layer with respect to
  the input, reshapes are necessary for broadcasting the
  element-wise product on the right axis
  """
  hidden = hidden * (1 - hidden)
  L = tf.expand_dims(hidden, 1) * tf.expand_dims(params, 0)
  # Compute the jacobian and average over the number of samples/minibatch
  L = tf.reduce_sum(tf.pow(L, 2)) / hidden.shape[0]
  return tf.reduce_mean(L)


def correntropy_regularize(x, sigma=1.):
  """
  Note
  ----
  origin implementation from seya:
  https://github.com/EderSantana/seya/blob/master/seya/regularizers.py
  Copyright (c) EderSantana
  """
  return -tf.reduce_sum(tf.reduce_mean(tf.exp(x**2 / sigma), axis=0)) / tf.sqrt(2 * np.pi * sigma)


def kl_gaussian(mu, logsigma,
                prior_mu=0., prior_logsigma=0.):
  """ KL-divergence between two gaussians.
  Useful for Variational AutoEncoders. Use this as an activation regularizer

  For taking kl_gaussian as variational regularization, you can take mean of
  the return matrix

  Parameters:
  -----------
  mean, logsigma: parameters of the input distributions
  prior_mean, prior_logsigma: paramaters of the desired distribution (note the
      log on logsigma)


  Return
  ------
  matrix: (n_samples, n_features)

  Note
  ----
  origin implementation from:
  https://github.com/Philip-Bachman/ICML-2015/blob/master/LogPDFs.py
  Copyright (c) Philip Bachman
  """
  if is_number(prior_mu):
    prior_mu = tf.convert_to_tensor(prior_mu, name='prior_mu',
        dtype=mu.dtype.base_dtype)
  if is_number(prior_logsigma):
    prior_logsigma = tf.convert_to_tensor(
        prior_logsigma, name='prior_logsigma',
        dtype=logsigma.dtype.base_dtype)
  gauss_klds = 0.5 * (2 * (prior_logsigma - logsigma) +
          (tf.exp(2 * logsigma) / tf.exp(2 * prior_logsigma)) +
          (tf.pow((mu - prior_mu), 2.0) / tf.exp(2 * prior_logsigma)) - 1.0)
  return gauss_klds



from __future__ import print_function, division, absolute_import

import tensorflow as tf
from tensorflow.python.keras.layers import Layer
from tensorflow_probability.python.distributions import Distribution
from tensorflow_probability.python.layers.internal import distribution_tensor_coercible as dtc

class ReduceMean(Layer):
  """ ReduceMean """

  def __init__(self, axis=None, keepdims=None, **kwargs):
    super(ReduceMean, self).__init__(**kwargs)
    self.axis = axis
    self.keepdims = keepdims

  def call(self, x):
    return [tf.reduce_mean(i, axis=self.axis, keepdims=self.keepdims) for i in x] \
    if isinstance(x, (tuple, list)) else \
    tf.reduce_mean(x, axis=self.axis, keepdims=self.keepdims)

class Moments(Layer):
  """ Moments """

  def __init__(self, mean=True, variance=True, **kwargs):
    super(Moments, self).__init__(**kwargs)
    self.mean = bool(mean)
    self.variance = bool(variance)
    assert self.mean or self.variance, "This layer must return mean or variance"

  def call(self, x):
    assert isinstance(x, Distribution), \
    "Input to this layer must be instance of tensorflow_probability Distribution"
    outputs = []
    if self.mean:
      outputs.append(x.mean())
    if self.variance:
      outputs.append(x.variance())
    return outputs[0] if len(outputs) == 1 else tuple(outputs)

  def compute_output_shape(self, input_shape):
    return [input_shape, input_shape] if self.mean and self.variance else input_shape

class Stddev(Layer):

  def __init__(self, **kwargs):
    super(Stddev, self).__init__(**kwargs)

  def call(self, x):
    assert isinstance(x, Distribution), \
    "Input to this layer must be instance of tensorflow_probability Distribution"
    return x.stddev()

  def compute_output_shape(self, input_shape):
    return input_shape

class GetAttr(Layer):
  """ GetAttr """

  def __init__(self, attr_name, convert_to_tensor_fn=Distribution.sample, **kwargs):
    super(GetAttr, self).__init__(**kwargs)
    self.attr_name = str(attr_name)
    if isinstance(convert_to_tensor_fn, property):
      convert_to_tensor_fn = convert_to_tensor_fn.fget
    self.convert_to_tensor_fn = convert_to_tensor_fn

  def call(self, x):
    attrs = self.attr_name.split('.')
    for a in attrs:
      x = getattr(x, a)
    # special case a distribution is returned
    if isinstance(x, Distribution) and not isinstance(x, dtc._TensorCoercible):
      dist = dtc._TensorCoercible(
          distribution=x,
          convert_to_tensor_fn=self.convert_to_tensor_fn)
      value = tf.convert_to_tensor(value=dist)
      value._tfp_distribution = dist
      dist.shape = value.shape
      dist.get_shape = value.get_shape
      x = dist
    return x



# -*- coding: utf-8 -*-
# Natural Language Toolkit: TextGrid analysis
#
# Copyright (C) 2001-2011 NLTK Project
# Author: Margaret Mitchell <itallow@gmail.com>
#         Steven Bird <sb@csse.unimelb.edu.au> (revisions)
# URL: <http://www.nltk.org>
# For license information, see LICENSE.TXT
#

"""
Tools for reading TextGrid files, the format used by Praat.

Module contents
===============

The textgrid corpus reader provides 4 data items and 1 function
for each textgrid file.  For each tier in the file, the reader
provides 10 data items and 2 functions.

For the full textgrid file:

  - size
    The number of tiers in the file.

  - xmin
    First marked time of the file.

  - xmax
    Last marked time of the file.

  - t_time
    xmax - xmin.

  - text_type
    The style of TextGrid format:
        - ooTextFile:  Organized by tier.
        - ChronTextFile:  Organized by time.
        - OldooTextFile:  Similar to ooTextFile.

  - to_chron()
    Convert given file to a ChronTextFile format.

  - to_oo()
    Convert given file to an ooTextFile format.

For each tier:

  - text_type
    The style of TextGrid format, as above.

  - classid
    The style of transcription on this tier:
        - IntervalTier:  Transcription is marked as intervals.
        - TextTier:  Transcription is marked as single points.

  - nameid
    The name of the tier.

  - xmin
    First marked time of the tier.

  - xmax
    Last marked time of the tier.

  - size
    Number of entries in the tier.

  - transcript
    The raw transcript for the tier.

  - simple_transcript
    The transcript formatted as a list of tuples: (time1, time2, utterance).

  - tier_info
    List of (classid, nameid, xmin, xmax, size, transcript).

  - min_max()
    A tuple of (xmin, xmax).

  - time(non_speech_marker)
    Returns the utterance time of a given tier.
    Excludes entries that begin with a non-speech marker.

"""

# needs more cleanup, subclassing, epydoc docstrings
from __future__ import print_function
import sys
import os
import re
from odin.utils import is_fileobj

TEXTTIER = "TextTier"
INTERVALTIER = "IntervalTier"

OOTEXTFILE = re.compile(r"""(?x)
            xmin\ =\ (.*)[\r\n]+
            xmax\ =\ (.*)[\r\n]+
            [\s\S]+?size\ =\ (.*)[\r\n]+
""")

CHRONTEXTFILE = re.compile(r"""(?x)
            [\r\n]+(\S+)\
            (\S+)\ +!\ Time\ domain.\ *[\r\n]+
            (\S+)\ +!\ Number\ of\ tiers.\ *[\r\n]+"
""")

OLDOOTEXTFILE = re.compile(r"""(?x)
            [\r\n]+(\S+)
            [\r\n]+(\S+)
            [\r\n]+.+[\r\n]+(\S+)
""")


#################################################################
# TextGrid Class
#################################################################
class TextGrid(object):

  """
  Class to manipulate the TextGrid format used by Praat.
  Separates each tier within this file into its own Tier
  object.  Each TextGrid object has
  a number of tiers (size), xmin, xmax, a text type to help
  with the different styles of TextGrid format, and tiers with their
  own attributes.
  """

  def __init__(self, read_file):
    """
    Takes open read file as input, initializes attributes
    of the TextGrid file.
    @type read_file: An open TextGrid file, mode "r".
    @param size:  Number of tiers.
    @param xmin: xmin.
    @param xmax: xmax.
    @param t_time:  Total time of TextGrid file.
    @param text_type:  TextGrid format.
    @type tiers:  A list of tier objects.
    """
    if isinstance(read_file, str) and os.path.isfile(read_file):
      read_file = open(read_file, 'r')
    if is_fileobj(read_file):
      tmp = read_file.read()
      read_file.close()
      read_file = tmp

    # clean the text little bit
    try:
      self.read_file = read_file.decode('utf-8')
    except Exception:
      self.read_file = read_file.decode('utf-16')

    self.size = 0
    self.xmin = 0
    self.xmax = 0
    self.t_time = 0
    self.text_type = self._check_type()
    self.tiers = self._find_tiers()

  def __iter__(self):
    for tier in self.tiers:
      yield tier

  def __getitem__(self, key):
    return self.tiers.__getitem__(key)

  def next(self):
    if self.idx == (self.size - 1):
      raise StopIteration
    self.idx += 1
    return self.tiers[self.idx]

  @staticmethod
  def load(file):
    """
    @param file: a file in TextGrid format
    """

    return TextGrid(open(file).read())

  def _load_tiers(self, header):
    """
    Iterates over each tier and grabs tier information.
    """

    tiers = []
    if self.text_type == "ChronTextFile":
      m = re.compile(header)
      tier_headers = m.findall(self.read_file)
      tier_re = " \d+.?\d* \d+.?\d*[\r\n]+\"[^\"]*\""
      for i in range(0, self.size):
        tier_info = [tier_headers[i]] + \
        re.findall(str(i + 1) + tier_re, self.read_file)
        tier_info = "\n".join(tier_info)
        tiers.append(Tier(tier_info, self.text_type, self.t_time))
      return tiers

    tier_re = header + "[\s\S]+?(?=" + header + "|$$)"
    m = re.compile(tier_re)
    tier_iter = m.finditer(self.read_file)
    for iterator in tier_iter:
      (begin, end) = iterator.span()
      tier_info = self.read_file[begin:end]
      tiers.append(Tier(tier_info, self.text_type, self.t_time))
    return tiers

  def _check_type(self):
    """
    Figures out the TextGrid format.
    """

    m = re.match(r"(.*)[\r\n](.*)[\r\n](.*)[\r\n](.*)", self.read_file)
    try:
      type_id = m.group(1).strip()
    except AttributeError:
      raise TypeError("Cannot read file -- try TextGrid.load()")
    xmin = m.group(4)
    if type_id == "File type = \"ooTextFile\"":
      if "xmin" not in xmin:
        text_type = "OldooTextFile"
      else:
        text_type = "ooTextFile"
    elif type_id == "\"Praat chronological TextGrid text file\"":
      text_type = "ChronTextFile"
    else:
      raise TypeError("Unknown format '(%s)'", (type_id))
    return text_type

  def _find_tiers(self):
    """
    Splits the textgrid file into substrings corresponding to tiers.
    """

    if self.text_type == "ooTextFile":
      m = OOTEXTFILE
      header = " +item \["
    elif self.text_type == "ChronTextFile":
      m = CHRONTEXTFILE
      header = "\"\S+\" \".*\" \d+\.?\d* \d+\.?\d*"
    elif self.text_type == "OldooTextFile":
      m = OLDOOTEXTFILE
      header = "\".*\"[\r\n]+\".*\""

    file_info = m.findall(self.read_file)[0]
    self.xmin = float(file_info[0])
    self.xmax = float(file_info[1])
    self.t_time = self.xmax - self.xmin
    self.size = int(file_info[2])
    tiers = self._load_tiers(header)
    return tiers

  def to_chron(self):
    """
    @return:  String in Chronological TextGrid file format.
    """

    chron_file = ""
    chron_file += "\"Praat chronological TextGrid text file\"\n"
    chron_file += str(self.xmin) + " " + str(self.xmax)
    chron_file += "   ! Time domain.\n"
    chron_file += str(self.size) + "   ! Number of tiers.\n"
    for tier in self.tiers:
      idx = (self.tiers.index(tier)) + 1
      tier_header = "\"" + tier.classid + "\" \"" \
                    + tier.nameid + "\" " + str(tier.xmin) \
                    + " " + str(tier.xmax)
      chron_file += tier_header + "\n"
      transcript = tier.simple_transcript
      for (xmin, xmax, utt) in transcript:
        chron_file += str(idx) + " " + str(xmin)
        chron_file += " " + str(xmax) + "\n"
        chron_file += "\"" + utt + "\"\n"
    return chron_file

  def to_oo(self):
    """
    @return:  A string in OoTextGrid file format.
    """

    oo_file = ""
    oo_file += "File type = \"ooTextFile\"\n"
    oo_file += "Object class = \"TextGrid\"\n\n"
    oo_file += "xmin = ", self.xmin, "\n"
    oo_file += "xmax = ", self.xmax, "\n"
    oo_file += "tiers? <exists>\n"
    oo_file += "size = ", self.size, "\n"
    oo_file += "item []:\n"
    for i in range(len(self.tiers)):
      oo_file += "%4s%s [%s]" % ("", "item", i + 1)
      _curr_tier = self.tiers[i]
      for (x, y) in _curr_tier.header:
        oo_file += "%8s%s = \"%s\"" % ("", x, y)
      if _curr_tier.classid != TEXTTIER:
        for (xmin, xmax, text) in _curr_tier.simple_transcript:
          oo_file += "%12s%s = %s" % ("", "xmin", xmin)
          oo_file += "%12s%s = %s" % ("", "xmax", xmax)
          oo_file += "%12s%s = \"%s\"" % ("", "text", text)
      else:
        for (time, mark) in _curr_tier.simple_transcript:
          oo_file += "%12s%s = %s" % ("", "time", time)
          oo_file += "%12s%s = %s" % ("", "mark", mark)
    return oo_file


#################################################################
# Tier Class
#################################################################
class Tier(object):

  """
  A container for each tier.
  """

  def __init__(self, tier, text_type, t_time):
    """
    Initializes attributes of the tier: class, name, xmin, xmax
    size, transcript, total time.
    Utilizes text_type to guide how to parse the file.
    @type tier: a tier object; single item in the TextGrid list.
    @param text_type:  TextGrid format
    @param t_time:  Total time of TextGrid file.
    @param classid:  Type of tier (point or interval).
    @param nameid:  Name of tier.
    @param xmin:  xmin of the tier.
    @param xmax:  xmax of the tier.
    @param size:  Number of entries in the tier
    @param transcript:  The raw transcript for the tier.
    """

    self.tier = tier
    self.text_type = text_type
    self.t_time = t_time
    self.classid = ""
    self.nameid = ""
    self.xmin = 0
    self.xmax = 0
    self.size = 0
    self._transcript = ""
    self.tier_info = ""
    self._make_info()
    self.simple_transcript = self.make_simple_transcript()
    if self.classid != TEXTTIER:
      self.mark_type = "intervals"
    else:
      self.mark_type = "points"
      self.header = [("class", self.classid), ("name", self.nameid),
      ("xmin", self.xmin), ("xmax", self.xmax), ("size", self.size)]

  def __iter__(self):
    return self.simple_transcript.__iter__()

  def _make_info(self):
    """
    Figures out most attributes of the tier object:
    class, name, xmin, xmax, transcript.
    """

    trans = "([\S\s]*)"
    if self.text_type == "ChronTextFile":
      classid = "\"(.*)\" +"
      nameid = "\"(.*)\" +"
      xmin = "(\d+\.?\d*) +"
      xmax = "(\d+\.?\d*) *[\r\n]+"
      # No size values are given in the Chronological Text File format.
      self.size = None
      size = ""
    elif self.text_type == "ooTextFile":
      classid = " +class = \"(.*)\" *[\r\n]+"
      nameid = " +name = \"(.*)\" *[\r\n]+"
      xmin = " +xmin = (\d+\.?\d*) *[\r\n]+"
      xmax = " +xmax = (\d+\.?\d*) *[\r\n]+"
      size = " +\S+: size = (\d+) *[\r\n]+"
    elif self.text_type == "OldooTextFile":
      classid = "\"(.*)\" *[\r\n]+"
      nameid = "\"(.*)\" *[\r\n]+"
      xmin = "(\d+\.?\d*) *[\r\n]+"
      xmax = "(\d+\.?\d*) *[\r\n]+"
      size = "(\d+) *[\r\n]+"
    m = re.compile(classid + nameid + xmin + xmax + size + trans)
    self.tier_info = m.findall(self.tier)[0]
    self.classid = self.tier_info[0]
    self.nameid = self.tier_info[1]
    self.xmin = float(self.tier_info[2])
    self.xmax = float(self.tier_info[3])
    if self.size is not None:
      self.size = int(self.tier_info[4])
    self._transcript = self.tier_info[-1]

  def make_simple_transcript(self):
    """
    @return:  Transcript of the tier, in form [(start_time end_time label)]
    """

    if self.text_type == "ChronTextFile":
      trans_head = ""
      trans_xmin = " (\S+)"
      trans_xmax = " (\S+)[\r\n]+"
      trans_text = "\"([\S\s]*?)\""
    elif self.text_type == "ooTextFile":
      trans_head = " +\S+ \[\d+\]: *[\r\n]+"
      trans_xmin = " +\S+ = (\S+) *[\r\n]+"
      trans_xmax = " +\S+ = (\S+) *[\r\n]+"
      trans_text = " +\S+ = \"([^\"]*?)\""
    elif self.text_type == "OldooTextFile":
      trans_head = ""
      trans_xmin = "(.*)[\r\n]+"
      trans_xmax = "(.*)[\r\n]+"
      trans_text = "\"([\S\s]*?)\""
    if self.classid == TEXTTIER:
      trans_xmin = ""
    trans_m = re.compile(trans_head + trans_xmin + trans_xmax + trans_text)
    self.simple_transcript = trans_m.findall(self._transcript)
    return self.simple_transcript

  def transcript(self, timestamp):
    """
    Parameters
    ----------
    timestamp : list
        start, and end timestamp of given duration

    @return:  Transcript of the tier, as it appears in the file.
    """

    start_t = float(timestamp[0])
    end_t = float(timestamp[1])
    duration = (end_t - start_t) / 2
    for start, end, trans in self:
      start = float(start)
      end = float(end)
      overlap = max(0, min(end, end_t) - max(start, start_t))
      if overlap >= duration:
        return trans
    return "" # no transcription available

  def time(self, non_speech_char="."):
    """
    @return: Utterance time of a given tier.
    Screens out entries that begin with a non-speech marker.
    """

    total = 0.0
    if self.classid != TEXTTIER:
      for (time1, time2, utt) in self.simple_transcript:
        utt = utt.strip()
        if utt and not utt[0] == "." and len(utt) > 0:
          total += (float(time2) - float(time1))
    return total

  def tier_name(self):
    """
    @return:  Tier name of a given tier.
    """

    return self.nameid

  def classid(self):
    """
    @return:  Type of transcription on tier.
    """

    return self.classid

  def min_max(self):
    """
    @return:  (xmin, xmax) tuple for a given tier.
    """

    return (self.xmin, self.xmax)

  def __repr__(self):
    s = u"<%s \"%s\" (%.2f, %.2f) %.2f%%>" % \
    (self.classid, self.nameid, self.xmin, self.xmax, 100 * self.time() / self.t_time)
    return s

  def __str__(self):
    s = self.__repr__() + u"\n  " + u"\n  ".join(u' '.join(row) for row in self.simple_transcript)
    try:
      return s.encode('utf-8')
    except:
      return s.encode('utf-16')


def demo_TextGrid(demo_data):
  print("** Demo of the TextGrid class. **")

  fid = TextGrid(demo_data)
  print("Tiers:", fid.size)

  for i, tier in enumerate(fid):
    print("\n***")
    print("Tier:", i + 1)
    print(tier)


def demo():
  # Each demo demonstrates different TextGrid formats.
  print("Format 1")
  demo_TextGrid(demo_data1)
  print("\nFormat 2")
  demo_TextGrid(demo_data2)
  print("\nFormat 3")
  demo_TextGrid(demo_data3)


demo_data1 = """File type = "ooTextFile"
Object class = "TextGrid"

xmin = 0
xmax = 2045.144149659864
tiers? <exists>
size = 3
item []:
    item [1]:
        class = "IntervalTier"
        name = "utterances"
        xmin = 0
        xmax = 2045.144149659864
        intervals: size = 5
        intervals [1]:
            xmin = 0
            xmax = 2041.4217474125382
            text = ""
        intervals [2]:
            xmin = 2041.4217474125382
            xmax = 2041.968276643991
            text = "this"
        intervals [3]:
            xmin = 2041.968276643991
            xmax = 2042.5281632653062
            text = "is"
        intervals [4]:
            xmin = 2042.5281632653062
            xmax = 2044.0487352585324
            text = "a"
        intervals [5]:
            xmin = 2044.0487352585324
            xmax = 2045.144149659864
            text = "demo"
    item [2]:
        class = "TextTier"
        name = "notes"
        xmin = 0
        xmax = 2045.144149659864
        points: size = 3
        points [1]:
            time = 2041.4217474125382
            mark = ".begin_demo"
        points [2]:
            time = 2043.8338291031832
            mark = "voice gets quiet here"
        points [3]:
            time = 2045.144149659864
            mark = ".end_demo"
    item [3]:
        class = "IntervalTier"
        name = "phones"
        xmin = 0
        xmax = 2045.144149659864
        intervals: size = 12
        intervals [1]:
            xmin = 0
            xmax = 2041.4217474125382
            text = ""
        intervals [2]:
            xmin = 2041.4217474125382
            xmax = 2041.5438290324326
            text = "D"
        intervals [3]:
            xmin = 2041.5438290324326
            xmax = 2041.7321032910372
            text = "I"
        intervals [4]:
            xmin = 2041.7321032910372
            xmax = 2041.968276643991
            text = "s"
        intervals [5]:
            xmin = 2041.968276643991
            xmax = 2042.232189031843
            text = "I"
        intervals [6]:
            xmin = 2042.232189031843
            xmax = 2042.5281632653062
            text = "z"
        intervals [7]:
            xmin = 2042.5281632653062
            xmax = 2044.0487352585324
            text = "eI"
        intervals [8]:
            xmin = 2044.0487352585324
            xmax = 2044.2487352585324
            text = "dc"
        intervals [9]:
            xmin = 2044.2487352585324
            xmax = 2044.3102321849011
            text = "d"
        intervals [10]:
            xmin = 2044.3102321849011
            xmax = 2044.5748932104329
            text = "E"
        intervals [11]:
            xmin = 2044.5748932104329
            xmax = 2044.8329108578437
            text = "m"
        intervals [12]:
            xmin = 2044.8329108578437
            xmax = 2045.144149659864
            text = "oU"
"""

demo_data2 = """File type = "ooTextFile"
Object class = "TextGrid"

0
2.8
<exists>
2
"IntervalTier"
"utterances"
0
2.8
3
0
1.6229213249309031
""
1.6229213249309031
2.341428074708195
"demo"
2.341428074708195
2.8
""
"IntervalTier"
"phones"
0
2.8
6
0
1.6229213249309031
""
1.6229213249309031
1.6428291382019483
"dc"
1.6428291382019483
1.65372183721983721
"d"
1.65372183721983721
1.94372874328943728
"E"
1.94372874328943728
2.13821938291038210
"m"
2.13821938291038210
2.341428074708195
"oU"
2.341428074708195
2.8
""
"""

demo_data3 = """"Praat chronological TextGrid text file"
0 2.8   ! Time domain.
2   ! Number of tiers.
"IntervalTier" "utterances" 0 2.8
"IntervalTier" "utterances" 0 2.8
1 0 1.6229213249309031
""
2 0 1.6229213249309031
""
2 1.6229213249309031 1.6428291382019483
"dc"
2 1.6428291382019483 1.65372183721983721
"d"
2 1.65372183721983721 1.94372874328943728
"E"
2 1.94372874328943728 2.13821938291038210
"m"
2 2.13821938291038210 2.341428074708195
"oU"
1 1.6229213249309031 2.341428074708195
"demo"
1 2.341428074708195 2.8
""
2 2.341428074708195 2.8
""
"""

if __name__ == "__main__":
  demo()



from __future__ import print_function, division, absolute_import

import matplotlib
matplotlib.use('Agg')
from matplotlib import pyplot as plt

import os
os.environ['ODIN'] = 'gpu,float32,seed=12082518'
from collections import defaultdict

import numpy as np
import tensorflow as tf

from odin import backend as K, nnet as N, fuel as F, visual as V
from odin.stats import train_valid_test_split, freqcount, sampling_iter
from odin import training
from odin.ml import evaluate, fast_pca, PLDA, Scorer
from odin.utils import (Progbar, unique_labels, as_tuple_of_shape, stdio,
                        ctext, args_parse, get_formatted_datetime,
                        batching)

from utils import (prepare_data, make_dnn_prediction, visualize_latent_space,
                   get_exp_path)

args = args_parse(descriptions=[
    ('-feat', 'Input feature for training', None, 'mspec24'),
    ('-task', 'gender, age, dialect, speaker, digit', None, 'gender'),
    ('-batch', 'batch size', None, 32),
    ('-epoch', 'Number of training epoch', None, 12),
    ('--retrain', "deleted trained model, and re-train everything", None, False)
])
# ===========================================================================
# Const
# ===========================================================================
EXP_DIR, MODEL_PATH, LOG_PATH = get_exp_path('tvec', args, override=args.retrain)
stdio(LOG_PATH)
# ====== load data feeder ====== #
(train, valid,
 X_test_name, X_test_true, X_test_data,
 labels) = prepare_data(feat=args.feat, label=args.task)
n_classes = len(labels)
# ===========================================================================
# Create model
# ===========================================================================
inputs = [K.placeholder(shape=(None,) + shape[1:], dtype='float32', name='input%d' % i)
          for i, shape in enumerate(as_tuple_of_shape(train.shape))]
X = inputs[0]
y = inputs[1]
print("Inputs:", ctext(inputs, 'cyan'))
# ====== create the networks ====== #
with N.args_scope(
    [('Conv', 'Dense'), dict(b_init=None, activation=K.linear, pad='same')],
        ['BatchNorm', dict(activation=K.relu)]):
  f = N.Sequence([
      N.Dimshuffle(pattern=(0, 1, 2, 'x')),

      N.Conv(num_filters=32, filter_size=(9, 7)), N.BatchNorm(),
      N.Pool(pool_size=(3, 2), strides=2),
      N.Conv(num_filters=64, filter_size=(5, 3)), N.BatchNorm(),
      N.Pool(pool_size=(3, 1), strides=(2, 1), name='PoolOutput1'),
      N.Conv(num_filters=64, filter_size=(5, 3)), N.BatchNorm(),
      N.Pool(pool_size=(3, 2), strides=(2, 2), name='PoolOutput2'),

      N.Flatten(outdim=2),

      N.Dense(512, name="LatentDense"), N.BatchNorm(),
      N.Dense(512), N.BatchNorm(),

      N.Dense(n_classes)
  ], debug=1)
# ====== create outputs ====== #
y_logit = f(X)
y_proba = tf.nn.softmax(y_logit)
z1 = K.ComputationGraph(y_proba).get(roles=N.Pool, scope='PoolOutput1',
                                     beginning_scope=False)[0]
z2 = K.ComputationGraph(y_proba).get(roles=N.Pool, scope='PoolOutput2',
                                     beginning_scope=False)[0]
z3 = K.ComputationGraph(y_proba).get(scope='LatentDense',
                                     beginning_scope=False)[0]
print('Latent space:', ctext([z1, z2, z3], 'cyan'))
# ====== create loss ====== #
ce = tf.losses.softmax_cross_entropy(onehot_labels=y, logits=y_logit)
acc = K.metrics.categorical_accuracy(y_true=y, y_pred=y_proba)
cm = K.metrics.confusion_matrix(y_true=y, y_pred=y_proba, labels=len(labels))
# ====== params and optimizing ====== #
updates = K.optimizers.Adam(lr=0.001).minimize(
    loss=ce, roles=[K.role.TrainableParameter],
    exclude_roles=[K.role.InitialState],
    verbose=True)
K.initialize_all_variables()
# ====== Functions ====== #
print('Building training functions ...')
f_train = K.function(inputs, [ce, acc, cm], updates=updates,
                     training=True)
print('Building testing functions ...')
f_score = K.function(inputs, [ce, acc, cm],
                    training=False)
print('Building predicting functions ...')
f_pred_proba = K.function(X, y_proba, training=False)
# Latent spaces
f_z1 = K.function(inputs=X, outputs=z1, training=False)
f_z2 = K.function(inputs=X, outputs=z2, training=False)
f_z3 = K.function(inputs=X, outputs=z3, training=False)
# ===========================================================================
# Training
# ===========================================================================
print('Start training ...')
task = training.MainLoop(batch_size=args.batch, seed=120825, shuffle_level=2,
                         allow_rollback=True, labels=labels)
task.set_checkpoint(MODEL_PATH, f)
task.set_callbacks([
    training.NaNDetector(),
    training.EarlyStopGeneralizationLoss('valid', ce,
                                         threshold=5, patience=5)
])
task.set_train_task(func=f_train, data=train,
                    epoch=args.epoch, name='train')
task.set_valid_task(func=f_score, data=valid,
                    freq=training.Timer(percentage=0.8),
                    name='valid')
task.run()
# ===========================================================================
# Prediction
# ===========================================================================
y_pred_proba, Z1_test, Z2_test, Z3_test = make_dnn_prediction(
    functions=[f_pred_proba, f_z1, f_z2, f_z3], X=X_test_data, title='TEST')
print("Test Latent:", Z1_test.shape, Z2_test.shape, Z3_test.shape)
y_pred = np.argmax(y_pred_proba, axis=-1)
evaluate(y_true=X_test_true, y_pred_proba=y_pred_proba, labels=labels,
         title="Test set (Deep prediction)",
         path=os.path.join(EXP_DIR, 'test_deep.pdf'))
# ====== make a streamline classifier ====== #
# training PLDA
Z3_train, y_train = make_dnn_prediction(f_z3, X=train, title="TRAIN")
print("Z3_train:", Z3_train.shape, y_train.shape)
Z3_valid, y_valid = make_dnn_prediction(f_z3, X=valid, title="VALID")
print("Z3_valid:", Z3_valid.shape, y_valid.shape)
plda = PLDA(n_phi=200, random_state=K.get_rng().randint(10e8),
            n_iter=12, labels=labels, verbose=0)
plda.fit(np.concatenate([Z3_train, Z3_valid], axis=0),
         np.concatenate([y_train, y_valid], axis=0))
y_pred_log_proba = plda.predict_log_proba(Z3_test)
evaluate(y_true=X_test_true, y_pred_log_proba=y_pred_log_proba, labels=labels,
         title="Test set (PLDA - Latent prediction)",
         path=os.path.join(EXP_DIR, 'test_latent.pdf'))
# ====== visualize ====== #
visualize_latent_space(X_org=X_test_data, X_latent=Z1_test,
                       name=X_test_name, labels=X_test_true,
                       title="latent1")
visualize_latent_space(X_org=X_test_data, X_latent=Z2_test,
                       name=X_test_name, labels=X_test_true,
                       title="latent2")
V.plot_save(os.path.join(EXP_DIR, 'latent.pdf'))



from numbers import Number

import numpy as np
import scipy as sp

from sklearn.mixture import GaussianMixture
from sklearn.utils.extmath import softmax

from odin.utils import as_tuple
from odin.ml.base import BaseEstimator, ClassifierMixin, Evaluable
from odin.ml.scoring import VectorNormalizer

class GMMclassifier(BaseEstimator, ClassifierMixin, Evaluable):
  """ GMMclassifier
  Parameters
  ----------
  strategy : str
    'ova' - one-vs-all, for each class represented as a GMM
    'all' - use a single GMM for all classes
  covariance_type : {'full', 'tied', 'diag', 'spherical'},
        defaults to 'full'.
    String describing the type of covariance parameters to use.
    Must be one of::
        'full' (each component has its own general covariance matrix),
        'tied' (all components share the same general covariance matrix),
        'diag' (each component has its own diagonal covariance matrix),
        'spherical' (each component has its own single variance).
  max_iter : int, defaults to 100.
    The number of EM iterations to perform.
  n_init : int, defaults to 1.
    The number of initializations to perform. The best results are kept.
  init_params : {'kmeans', 'random'}, defaults to 'kmeans'.
    The method used to initialize the weights, the means and the
    precisions.
    Must be one of::
        'kmeans' : responsibilities are initialized using kmeans.
        'random' : responsibilities are initialized randomly.
  n_components : {int, list of int}
    only used in case `strategy='ova'`, number of Gaussian components
    for each class
  """

  def __init__(self, strategy="ova", covariance_type='full',
               max_iter=100, n_init=1,
               init_params='kmeans', n_components=1,
               centering=False, wccn=False, unit_length=False,
               lda=False, concat=False, labels=None, seed=5218):
    super(GMMclassifier, self).__init__()
    self._strategy = str(strategy)
    self._n_components = int(n_components)
    self._covariance_type = str(covariance_type)
    self._max_iter = int(max_iter)
    self._n_init = int(n_init)
    self._init_params = str(init_params)
    self._seed = 5218 if not isinstance(seed, Number) else int(seed)
    # ====== default attribute ====== #
    self._labels = labels
    self._feat_dim = None
    self._gmm = None
    self._normalizer = VectorNormalizer(
        centering=centering, wccn=wccn, unit_length=unit_length,
        lda=lda, concat=concat)

  # ==================== Pickling ==================== #
  def __getstate__(self):
    if not self.is_fitted:
      raise RuntimeError("The GMMclassifier have not been fitted, "
                         "nothing to pickle!")
    return (self._strategy, self._n_components, self._covariance_type,
            self._max_iter, self._n_init, self._init_params,
            self._labels, self._feat_dim, self._gmm,
            self._normalizer)

  def __setstate__(self, states):
    (self._strategy, self._n_components, self._covariance_type,
     self._max_iter, self._n_init, self._init_params,
     self._labels, self._feat_dim, self._gmm,
     self._normalizer) = states

  # ==================== Properties ==================== #
  @property
  def is_fitted(self):
    return self._gmm is not None

  @property
  def feat_dim(self):
    return self._feat_dim

  @property
  def labels(self):
    return self._labels

  @property
  def nb_classes(self):
    return len(self.labels)

  # ==================== Helpers ==================== #
  def initialize(self, X, y=None):
    if isinstance(X, (tuple, list)):
      X = np.array(X)
    elif not isinstance(X, np.ndarray):
      X = X[:]
    if isinstance(y, (tuple, list)):
      y = np.array(y)
    elif y is not None and not isinstance(y, np.ndarray):
      y = y[:]
    # ====== check dimensions ====== #
    feat_dim = X.shape[1]
    if self._feat_dim is None:
      self._feat_dim = feat_dim
    if y is not None:
      classes = np.unique(y)
      if self._labels is None:
        self._labels = classes
    else:
      classes = self.labels
    # ====== exception ====== #
    if self.feat_dim != feat_dim:
      raise ValueError("Initialized with `feat_dim`=%d, given data with %d "
                       "dimensions" % (self.feat_dim, feat_dim))
    if self.nb_classes != len(classes):
      raise ValueError("Initialized with `nb_classes`=%d, given data with %d "
                       "classes" % (self.nb_classes, len(classes)))
    # ====== normalizing ====== #
    if not self._normalizer.is_fitted:
      self._normalizer.fit(X, y)
    X = self._normalizer.transform(X)
    # ====== initialize GMMs ====== #
    if self._gmm is None:
      if self._strategy == 'ova':
        self._gmm = []
        rand = np.random.RandomState(seed=self._seed)
        for n_components in as_tuple(self._n_components, t=int, N=self.nb_classes):
          gmm = GaussianMixture(n_components=n_components,
            covariance_type=self._covariance_type, max_iter=self._max_iter,
            n_init=self._n_init, init_params=self._init_params,
            random_state=rand.randint(0, 10e8))
          self._gmm.append(gmm)
      elif self._strategy == 'all':
        gmm = GaussianMixture(n_components=self.nb_classes,
            covariance_type=self._covariance_type, max_iter=self._max_iter,
            n_init=self._n_init, init_params=self._init_params,
            means_init=np.array([X[y == clz].mean(axis=0) for clz in np.unique(y)]),
            random_state=self._seed)
        self._gmm = gmm
      else:
        raise ValueError("No support for `strategy`=%s" % self._strategy)
    # ====== return ====== #
    if y is None:
      return X
    return X, y

  # ==================== Sklearn ==================== #
  def fit(self, X, y):
    X, y = self.initialize(X, y)
    classes = np.unique(y)
    if self._strategy == 'ova':
      for i, (clz, gmm) in enumerate(zip(classes, self._gmm)):
        X_cls = X[y == clz]
        gmm.fit(X_cls)
    elif self._strategy == 'all':
      self._gmm.fit(X)

  def score_samples(self, X):
    X = self.initialize(X)
    if self._strategy == 'ova':
      scores = np.concatenate([gmm.score_samples(X)[:, None]
                               for k, gmm in enumerate(self._gmm)],
                              axis=-1)
    elif self._strategy == 'all':
      scores = self._gmm.predict_proba(X)
    return scores

  def predict(self, X):
    return np.argmax(self.score_samples(X), axis=-1)

  def predict_proba(self, X):
    if self._strategy == 'ova':
      scores = self.score_samples(X)
      smin = np.min(scores, axis=-1, keepdims=True)
      smax = np.max(scores, axis=-1, keepdims=True)
      scores = (scores - smin) / (smax - smin)
      proba = softmax(scores)
    elif self._strategy == 'all':
      proba = self.score_samples(X)
    return proba

  def predict_log_proba(self, X):
    if self._strategy == 'ova':
      return self.score_samples(X)
    elif self._strategy == 'all':
      return np.log(self.predict_proba(X))



# ===========================================================================
# This module is created based on the code from libraries: Lasagne
# Original work Copyright (c) 2014-2015 Lasagne contributors
# Modified work Copyright 2016-2017 TrungNT
# ===========================================================================
from __future__ import print_function, division, absolute_import

import numbers
from six import add_metaclass, string_types
from abc import ABCMeta, abstractproperty
from collections import OrderedDict, defaultdict

import numpy as np
import tensorflow as tf

from odin.autoconfig import CONFIG, get_session
from odin.utils import as_tuple, is_number, uuid, ctext
from odin.utils.cache_utils import cache_memory
from odin.backend.role import (add_roles, Auxiliary, LearningRate, OptimizerHyperParameter,
                               GradientsNorm, GraidentsClippingNorm, GraidentsClippingValue,
                               has_roles, get_roles,
                               Role, Parameter, Weight, Bias, TrainableParameter,
                               Variable as _Variable, OptimizerVariable)
from odin.backend.helpers import (is_tensor, get_value, ComputationGraph, is_variable,
                                  get_all_variables, set_value)

floatX = CONFIG.floatX

__all__ = [
    "Optimizer",
    "SGD",
    "RMSProp",
    "Adadelta",
    "Adam",
    "Adamax",
    "Nadam",
    "Adagrad"
]
# ===========================================================================
# Helper methods
# ===========================================================================
def _as_variable(x, name, roles=None):
  # nothing to do
  if x is None:
    return None
  # create variable
  if not is_tensor(x):
    x = tf.Variable(x, dtype=floatX, name=name)
    get_session().run(x.initializer)
  return add_roles(x, roles)

# ===========================================================================
# Optimizer
# ===========================================================================
@add_metaclass(ABCMeta)
class Optimizer(object):

  """
  Parameters
  ----------
  lr: float, variable
      learning rate
  decay_steps: A scalar `int32` or `int64` `Tensor` or a Python number.
    Must be positive. (e.g. decay every 100000 steps with a base of 0.96)
  decay_rate: A scalar `float32` or `float64` `Tensor` or a
    Python number.  The decay rate.
  staircase: boolean.
      If `True` decay the learning rate at discrete intervals
  clipnorm: float >= 0. Gradients will be clipped
      when their L2 norm exceeds this value.
  clipvalue: float >= 0. Gradients will be clipped
      when their absolute value exceeds this value.
  clip_alg: str
      clipping algorithm for the gradients, one of the followings are
      accepted: "norm", "total_norm", "avg_norm"

  Note
  ----
  decayed_learning_rate = learning_rate *
                      decay_rate ^ (global_step / decay_steps)
  """

  def __init__(self, lr, decay_steps=None, decay_rate=0.96, staircase=True,
               clipnorm=None, clipvalue=None, clip_alg='total_norm',
               name=None):
    if name is None:
      name = self.__class__.__name__ + '_' + str(uuid(length=4))
    elif not isinstance(name, string_types):
      name = str(name)
    self._name = str(name)
    self.staircase = bool(staircase)
    with tf.variable_scope(self._name):
      self._lr = _as_variable(lr, name='learning_rate', roles=LearningRate)
      self._lr_decay = None
      self._step = tf.Variable(0., dtype=floatX,
          name="%s_step" % self.__class__.__name__)
      self.decay_steps = decay_steps
      self.decay_rate = decay_rate

      if clipnorm is not None:
        if (clipnorm if is_number(clipnorm) else get_value(clipnorm)) <= 0:
          raise ValueError('`clipnorm` value must greater than 0.')
      self.clipnorm = _as_variable(clipnorm, name="clip_norm",
          roles=GraidentsClippingNorm)

      if clipvalue is not None:
        if (clipvalue if is_number(clipvalue) else get_value(clipvalue)) <= 0:
          raise ValueError('`clipvalue` value must greater than 0.')
      self.clipvalue = _as_variable(clipvalue, name="clip_value",
          roles=GraidentsClippingValue)
    # ====== internal states values ====== #
    clip_alg = str(clip_alg).strip().lower()
    if clip_alg not in ('total_norm', 'norm', 'avg_norm'):
      raise ValueError("clip_arg must be one of the following: "
          "'norm', 'total_norm', 'avg_norm'")
    self._norm = 0.
    self.clip_alg = clip_alg
    self._algorithm = None
    self._is_initialized = False

  @property
  def name(self):
    return self._name

  @property
  def step(self):
    return self._step

  @property
  def lr(self):
    """ Return exponential_decay LR if the decay_steps is given
    otherwise, return the lr variable.
    """
    if self.decay_steps is not None:
      if self._lr_decay is None:
        self._lr_decay = tf.train.exponential_decay(self._lr,
            self._step, self.decay_steps, self.decay_rate,
            staircase=self.staircase)
      return self._lr_decay
    else:
      return self._lr

  @property
  def algorithm(self):
    return self._algorithm

  @property
  def norm(self):
    """Return L2-norm value of all gradients """
    return self._norm

  def minimize(self, loss, var_list=None,
               roles=[TrainableParameter], exclude_roles=[],
               reuse_scope=None, verbose=False):
    """
    Parameters
    ----------
    loss : {Tensor, list of Tensor}
      the loss for minimize
    var_list : {None, list of tensorflow.Variable}
      if None, automatically selecting the Variables, by getting
      all Variables associated with `loss` and using given
      `roles` and `exclude_roles`
    roles : {None, list of odin.backend.role}
      in case of `var_list=None`, all Variables with given role
      in this list will be selected for calculating gradients.
    exclude_roles : {None, list of odin.backend.role}
      all Variables have role in this list will be ignored.
    reuse_scope : {None, string}
      if provided, searching for all `OptimizerVariable` within
      given scope, and set the variables of this optimizer to
      the desire values
    verbose : bool (default: False)
      if True, print out all found variables and their roles
    """
    if exclude_roles is None:
      exclude_roles = []
    # ====== get all relevant variables ====== #
    if var_list is not None:
      all_variables = as_tuple(var_list, t=is_variable)
    else:
      all_variables = ComputationGraph(loss).variables
    # ====== filtering by Roles ====== #
    trainable = [v for v in all_variables
                 if has_roles(v, roles=roles)]
    if isinstance(exclude_roles, Role) or len(exclude_roles) > 0:
      trainable = [v for v in trainable
                   if not has_roles(v, roles=exclude_roles)]
    # ====== filtering by dtype ====== #
    # remove all boolean and string dtype
    trainable = [v
                 for v in trainable
                 if 'int' in str(v.dtype) or 'float' in str(v.dtype)]
    # ====== verbose ====== #
    if bool(verbose):
      # print loss first
      print("Loss:", ctext(str(loss), 'yellow'))
      # organize variable by role
      role_vars = defaultdict(list)
      for var in trainable:
        # remove all tensorflow collection name string
        roles = [r
                 for r in get_roles(var, return_string=False)
                 if not isinstance(r, string_types)]
        # only select highest role in the hierarchy
        high_roles = []
        for r1 in roles:
          found_ancester = False
          for r2 in roles:
            if r1 != r2 and issubclass(r2, r1):
              found_ancester = True
          if not found_ancester:
            high_roles.append(r1)
        # append to mapping role -> var_list
        for r in roles:
          role_vars[r.__name__].append(var)
      # print debug info (this is ugly but look nice and fun)
      if len(role_vars) > 0:
        max_name_length = str(max(
            max(len(i.name) for i in v)
            for v in role_vars.values()))
        max_shape_length = str(max(
            max(len(str(i.shape.as_list())) for i in v)
            for v in role_vars.values()))

        for role, var_list in sorted(role_vars.items(),
                                     key=lambda x: str(x[0])):
          print('Role:', ctext(role, 'yellow'))
          for var in sorted(var_list, key=lambda x: x.name):
            print(' ',
                  ('%-' + max_name_length + 's') % var.name.replace('/', ' '),
                  ctext(('%-' + max_shape_length + 's') % var.shape.as_list(), 'magenta'),
                  ctext(var.dtype.base_dtype.name, 'cyan'))
    # ====== get the updates ====== #
    updates = self.get_updates(loss_or_grads=loss, params=trainable)
    # with tf.variable_scope(self.name):
    #   updates = self.algorithm.minimize(loss=loss, global_step=self._step,
    #                                     var_list=trainable)
    # ====== re-use variable from previous scope ====== #
    if isinstance(reuse_scope, string_types):
      old_variables = [v for v in get_all_variables(scope=reuse_scope)
                       if has_roles(v, OptimizerVariable)]
      old_variables = sorted(old_variables,
                             key=lambda x: x.name.replace(reuse_scope, ''))
      old_varname = [v.name.replace(reuse_scope, '') for v in old_variables]
      # the hyper-parameters may not be included in `old_variables`
      # so they are needed to be excluded in new_variables
      new_variables = [v for v in get_all_variables(scope=self.name)
                       if has_roles(v, OptimizerVariable)]
      new_variables = sorted(new_variables,
                             key=lambda x: x.name.replace(self.name, ''))
      new_variables = [v for v in new_variables
                       if v.name.replace(self.name, '') in old_varname]
      assert len(old_variables) == len(new_variables), \
      "Number of variables in scope:'%s' is %d mismatch scope:'%s' with %d variables"\
      % (reuse_scope, len(old_variables), self.name, len(new_variables))
      for v_old, v_new in zip(old_variables, new_variables):
        assert v_new.shape == v_old.shape
        set_value(v_new, get_value(v_old), return_ops=False)
    return updates

  @cache_memory
  def get_updates(self, loss_or_grads, params):
    grads_vars = self.get_gradients(loss_or_grads, params)
    with tf.variable_scope(self.name, reuse=tf.AUTO_REUSE) as scope:
      scope_name = scope.name
      updates = self.algorithm.apply_gradients(grads_vars,
                                               global_step=self._step)
    for v in get_all_variables(scope=scope_name):
      add_roles(v, roles=OptimizerVariable)
    return updates

  def __call__(self, loss_or_grads, params):
    return self.get_updates(loss_or_grads, params)

  def grad(self, loss_or_grads, params):
    return self.get_gradients(loss_or_grads=loss_or_grads,
                              params=params)

  @cache_memory
  def get_gradients(self, loss_or_grads, params):
    """
    Note
    ----
    The returned gradients may contain None value
    """
    # check valid algorithm
    if self.algorithm is None or \
    not hasattr(self.algorithm, 'compute_gradients') or \
    not hasattr(self.algorithm, 'apply_gradients'):
      raise RuntimeError("Optimizer is None, or doesn't has attributes: "
                         "compute_gradients and apply_gradients.")
    with tf.variable_scope(self.name, reuse=tf.AUTO_REUSE) as scope:
      scope_name = scope.name
      # get the gradient
      grads_var = self.algorithm.compute_gradients(loss_or_grads,
                                                   var_list=params)
      grads_var = {g: v for g, v in grads_var if g is not None}
      grads = list(grads_var.keys())
      params = list(grads_var.values())
      # ====== clipnorm ====== #
      if self.clipnorm is not None:
        if self.clip_alg == 'norm':
          grads = [tf.clip_by_norm(g, self.clipnorm)
                   for g in grads]
        elif self.clip_alg == 'total_norm':
          grads, _ = tf.clip_by_global_norm(grads, self.clipnorm)
        elif self.clip_alg == 'avg_norm':
          grads = [tf.clip_by_average_norm(g, self.clipnorm)
                   for g in grads]
        else:
          raise ValueError("Unknown norm clipping algorithm: '%s'" % self.clip_alg)
      # ====== clipvalue ====== #
      if self.clipvalue is not None:
        grads = [tf.clip_by_value(g, -self.clipvalue, self.clipvalue)
                 for g in grads]
      # ====== get final norm value ====== #
      self._norm = add_roles(tf.global_norm(grads, name="GradientNorm"),
                             GradientsNorm)
    # ====== setting Optimizer roles ====== #
    for v in get_all_variables(scope=scope_name):
      add_roles(v, roles=OptimizerVariable)
    return [(g, p) for g, p in zip(grads, params)]

class SGD(Optimizer):

  """

  Parameters
  ----------
  momentum: float >= 0. None
      Parameter updates momentum. If momentum is None or 0,
      no momentum is not applied
  nesterov: boolean.
      Whether to apply Nesterov momentum.
  decay_steps: A scalar `int32` or `int64` `Tensor` or a Python number.
    Must be positive. (e.g. decay every 100000 steps with a base of 0.96)
  decay_rate: A scalar `float32` or `float64` `Tensor` or a
    Python number.  The decay rate.
  staircase: boolean.
      If `True` decay the learning rate at discrete intervals
  clipnorm: float >= 0. Gradients will be clipped
      when their L2 norm exceeds this value.
  clipvalue: float >= 0. Gradients will be clipped
      when their absolute value exceeds this value.
  clip_alg: str
      clipping algorithm for the gradients, one of the followings are
      accepted: "norm", "total_norm", "avg_norm"

  """

  def __init__(self, lr=0.01, momentum=0.9, nesterov=False,
               decay_steps=None, decay_rate=0.96, staircase=True,
               clipnorm=None, clipvalue=None,
               clip_alg='total_norm', name=None):
    super(SGD, self).__init__(lr=lr,
        decay_steps=decay_steps, decay_rate=decay_rate, staircase=staircase,
        clipnorm=clipnorm, clipvalue=clipvalue, clip_alg=clip_alg,
        name=name)
    with tf.variable_scope(self.name):
      # ====== momentum ====== #
      self.momentum = _as_variable(momentum, name='momentum',
                                   roles=OptimizerHyperParameter)
      self.nesterov = nesterov
      # ====== decay ====== #
      if self.momentum is not None:
        self._algorithm = tf.train.MomentumOptimizer(self.lr,
            self.momentum, use_nesterov=nesterov)
      else:
        self._algorithm = tf.train.GradientDescentOptimizer(
            learning_rate=self.lr)


class RMSProp(Optimizer):
  """RMSProp updates

  Scale learning rates by dividing with the moving average of the root mean
  squared (RMS) gradients. See [1]_ for further description.

  Parameters
  ----------
  lr : float or symbolic scalar
      The learning rate controlling the size of update steps
  rho : float or symbolic scalar
      Gradient moving average decay factor
  epsilon : float or symbolic scalar
      Small value added for numerical stability, `epsilon` might
      have huge impact on final performance.
  decay_steps: A scalar `int32` or `int64` `Tensor` or a Python number.
    Must be positive. (e.g. decay every 100000 steps with a base of 0.96)
  decay_rate: A scalar `float32` or `float64` `Tensor` or a
    Python number.  The decay rate.
  staircase: boolean.
      If `True` decay the learning rate at discrete intervals
  clipnorm: float >= 0. Gradients will be clipped
      when their L2 norm exceeds this value.
  clipvalue: float >= 0. Gradients will be clipped
      when their absolute value exceeds this value.
  clip_alg: str
      clipping algorithm for the gradients, one of the followings are
      accepted: "norm", "total_norm", "avg_norm"

  Returns
  -------
  OrderedDict
      A dictionary mapping each parameter to its update expression

  Notes
  -----
  `rho` should be between 0 and 1. A value of `rho` close to 1 will decay the
  moving average slowly and a value close to 0 will decay the moving average
  fast.

  Using the step size :math:`\\eta` and a decay factor :math:`\\rho` the
  learning rate :math:`\\eta_t` is calculated as:

  .. math::
     r_t &= \\rho r_{t-1} + (1-\\rho)*g^2\\\\
     \\eta_t &= \\frac{\\eta}{\\sqrt{r_t + \\epsilon}}

  References
  ----------
  .. [1] Tieleman, T. and Hinton, G. (2012):
         Neural Networks for Machine Learning, Lecture 6.5 - rmsprop.
         Coursera. http://www.youtube.com/watch?v=O3sxAc4hxZU (formula @5:20)
  """

  def __init__(self, lr=0.001, rho=0.9, momentum=0.0, epsilon=1e-10,
               decay_steps=None, decay_rate=0.96, staircase=True,
               clipnorm=None, clipvalue=None, clip_alg='total_norm',
               name=None):
    super(RMSProp, self).__init__(lr=lr,
        decay_steps=decay_steps, decay_rate=decay_rate, staircase=staircase,
        clipnorm=clipnorm, clipvalue=clipvalue, clip_alg=clip_alg,
        name=name)
    with tf.variable_scope(self.name):
      self.rho = _as_variable(rho, name='rho',
                              roles=OptimizerHyperParameter)
      self.momentum = _as_variable(momentum, name='momentum',
                              roles=OptimizerHyperParameter)
      self.epsilon = epsilon
      self._algorithm = tf.train.RMSPropOptimizer(self.lr,
          decay=self.rho, momentum=self.momentum, epsilon=self.epsilon,
          centered=False)


class Adadelta(Optimizer):
  """ Adadelta updates

  Scale learning rates by the ratio of accumulated gradients to accumulated
  updates, see [1]_ and notes for further description.

  Parameters
  ----------
  lr : float or symbolic scalar
      The learning rate controlling the size of update steps
  rho : float or symbolic scalar
      Squared gradient moving average decay factor
  epsilon : float or symbolic scalar
      Small value added for numerical stability, `epsilon` might
      have huge impact on final performance.
  decay_steps: A scalar `int32` or `int64` `Tensor` or a Python number.
    Must be positive. (e.g. decay every 100000 steps with a base of 0.96)
  decay_rate: A scalar `float32` or `float64` `Tensor` or a
    Python number.  The decay rate.
  staircase: boolean.
      If `True` decay the learning rate at discrete intervals
  clipnorm: float >= 0. Gradients will be clipped
      when their L2 norm exceeds this value.
  clipvalue: float >= 0. Gradients will be clipped
      when their absolute value exceeds this value.
  clip_alg: str
      clipping algorithm for the gradients, one of the followings are
      accepted: "norm", "total_norm", "avg_norm"

  Returns
  -------
  OrderedDict
      A dictionary mapping each parameter to its update expression

  Notes
  -----
  rho should be between 0 and 1. A value of rho close to 1 will decay the
  moving average slowly and a value close to 0 will decay the moving average
  fast.

  rho = 0.95 and epsilon=1e-6 are suggested in the paper and reported to
  work for multiple datasets (MNIST, speech).

  In the paper, no learning rate is considered (so learning_rate=1.0).
  Probably best to keep it at this value.
  epsilon is important for the very first update (so the numerator does
  not become 0).

  Using the step size eta and a decay factor rho the learning rate is
  calculated as:

  .. math::
     r_t &= \\rho r_{t-1} + (1-\\rho)*g^2\\\\
     \\eta_t &= \\eta \\frac{\\sqrt{s_{t-1} + \\epsilon}}
                           {\sqrt{r_t + \epsilon}}\\\\
     s_t &= \\rho s_{t-1} + (1-\\rho)*(\\eta_t*g)^2

  References
  ----------
  .. [1] Zeiler, M. D. (2012):
         ADADELTA: An Adaptive Learning Rate Method.
         arXiv Preprint arXiv:1212.5701.
  """

  def __init__(self, lr=1.0, rho=0.95, epsilon=1e-8,
               decay_steps=None, decay_rate=0.96, staircase=True,
               clipnorm=None, clipvalue=None, clip_alg='total_norm',
               name=None):
    super(Adadelta, self).__init__(lr=lr,
        decay_steps=decay_steps, decay_rate=decay_rate, staircase=staircase,
        clipnorm=clipnorm, clipvalue=clipvalue, clip_alg=clip_alg,
        name=name)
    with tf.variable_scope(self.name):
      self.rho = _as_variable(rho, name='rho',
                              roles=OptimizerHyperParameter)
      self.epsilon = epsilon
      self._algorithm = tf.train.AdadeltaOptimizer(
          learning_rate=self.lr, rho=self.rho,
          epsilon=self.epsilon)

class Adam(Optimizer):
  """Adam updates

  Adam updates implemented as in [1]_.

  Parameters
  ----------
  lr : float
      Learning rate
  beta1 : float
      Exponential decay rate for the first moment estimates.
  beta2 : float
      Exponential decay rate for the second moment estimates.
  epsilon : float or symbolic scalar
      Small value added for numerical stability, `epsilon` might
      have huge impact on final performance.
  decay_steps: A scalar `int32` or `int64` `Tensor` or a Python number.
    Must be positive. (e.g. decay every 100000 steps with a base of 0.96)
  decay_rate: A scalar `float32` or `float64` `Tensor` or a
    Python number.  The decay rate.
  staircase: boolean.
      If `True` decay the learning rate at discrete intervals
  clipnorm: float >= 0. Gradients will be clipped
      when their L2 norm exceeds this value.
  clipvalue: float >= 0. Gradients will be clipped
      when their absolute value exceeds this value.
  clip_alg: str
      clipping algorithm for the gradients, one of the followings are
      accepted: "norm", "total_norm", "avg_norm"

  Returns
  -------
  OrderedDict
      A dictionary mapping each parameter to its update expression

  Notes
  -----
  The paper [1]_ includes an additional hyperparameter lambda. This is only
  needed to prove convergence of the algorithm and has no practical use
  (personal communication with the authors), it is therefore omitted here.

  References
  ----------
  .. [1] Kingma, Diederik, and Jimmy Ba (2014):
         Adam: A Method for Stochastic Optimization.
         arXiv preprint arXiv:1412.6980.
  """

  def __init__(self, lr=0.001, beta1=0.9, beta2=0.999, epsilon=1e-8,
               decay_steps=None, decay_rate=0.96, staircase=True,
               clipnorm=None, clipvalue=None, clip_alg='total_norm',
               name=None):
    super(Adam, self).__init__(lr=lr,
        decay_steps=decay_steps, decay_rate=decay_rate, staircase=staircase,
        clipnorm=clipnorm, clipvalue=clipvalue, clip_alg=clip_alg,
        name=name)
    with tf.variable_scope(self.name):
      self.beta1 = _as_variable(beta1, name='beta1',
                                roles=OptimizerHyperParameter)
      self.beta2 = _as_variable(beta2, name='beta2',
                                roles=OptimizerHyperParameter)
      self.epsilon = epsilon
      self._algorithm = tf.train.AdamOptimizer(learning_rate=self.lr,
          beta1=self.beta1, beta2=self.beta2, epsilon=self.epsilon)

class Adamax(Optimizer):
  """Adamax updates

  Adamax updates implemented as in [1]_. This is a variant of of the Adam
  algorithm based on the infinity norm.

  Parameters
  ----------
  lr : float
      Learning rate
  beta1 : float
      Exponential decay rate for the first moment estimates.
  beta2 : float
      Exponential decay rate for the weighted infinity norm estimates.
  epsilon : float or symbolic scalar
      Small value added for numerical stability, `epsilon` might
      have huge impact on final performance.
  decay_steps: A scalar `int32` or `int64` `Tensor` or a Python number.
    Must be positive. (e.g. decay every 100000 steps with a base of 0.96)
  decay_rate: A scalar `float32` or `float64` `Tensor` or a
    Python number.  The decay rate.
  staircase: boolean.
      If `True` decay the learning rate at discrete intervals
  clipnorm: float >= 0. Gradients will be clipped
      when their L2 norm exceeds this value.
  clipvalue: float >= 0. Gradients will be clipped
      when their absolute value exceeds this value.
  clip_alg: str
      clipping algorithm for the gradients, one of the followings are
      accepted: "norm", "total_norm", "avg_norm"

  Returns
  -------
  OrderedDict
      A dictionary mapping each parameter to its update expression

  References
  ----------
  .. [1] Kingma, Diederik, and Jimmy Ba (2014):
         Adam: A Method for Stochastic Optimization.
         arXiv preprint arXiv:1412.6980.
  """

  def __init__(self, lr=0.002, beta_1=0.9, beta_2=0.999, epsilon=1e-8,
               clipnorm=None, clipvalue=None, clip_alg='total_norm',
               name=None):
    super(Adamax, self).__init__(lr, clipnorm, clipvalue, clip_alg=clip_alg,
                                 name=name)
    with tf.variable_scope(self.name):
      self.iterations = _as_variable(0, name='iterations', roles=Auxiliary)
      self.beta_1 = _as_variable(beta_1, name='beta_1',
                                 roles=OptimizerHyperParameter)
      self.beta_2 = _as_variable(beta_2, name='beta_2',
                                 roles=OptimizerHyperParameter)
      self.epsilon = epsilon
      raise NotImplementedError


class Nadam(Optimizer):
  """Nadam updates

  Nesterov Adam optimizer: Much like Adam is essentially RMSprop with momentum,
  Nadam is Adam RMSprop with Nesterov momentum.

  Parameters
  ----------
  lr : float
      Learning rate
  beta1 : float
      Exponential decay rate for the first moment estimates.
  beta2 : float
      Exponential decay rate for the second moment estimates.
  epsilon : float or symbolic scalar
      Small value added for numerical stability, `epsilon` might
      have huge impact on final performance.
  decay_steps: A scalar `int32` or `int64` `Tensor` or a Python number.
    Must be positive. (e.g. decay every 100000 steps with a base of 0.96)
  decay_rate: A scalar `float32` or `float64` `Tensor` or a
    Python number.  The decay rate.
  staircase: boolean.
      If `True` decay the learning rate at discrete intervals
  clipnorm: float >= 0. Gradients will be clipped
      when their L2 norm exceeds this value.
  clipvalue: float >= 0. Gradients will be clipped
      when their absolute value exceeds this value.
  clip_alg: str
      clipping algorithm for the gradients, one of the followings are
      accepted: "norm", "total_norm", "avg_norm"

  Returns
  -------
  OrderedDict
      A dictionary mapping each parameter to its update expression

  Notes
  -----
  The paper [1]_ includes an additional hyperparameter lambda. This is only
  needed to prove convergence of the algorithm and has no practical use
  (personal communication with the authors), it is therefore omitted here.

  References
  ----------
  .. [1] [Nadam report](http://cs229.stanford.edu/proj2015/054_report.pdf)
  .. [2] [On the importance of initialization and momentum in deep learning](http://www.cs.toronto.edu/~fritz/absps/momentum.pdf)

  """

  def __init__(self, lr=0.002, beta_1=0.9, beta_2=0.999, epsilon=1e-8,
               schedule_decay=0.004, clipnorm=None, clipvalue=None,
               clip_alg='total_norm', name=None):
    super(Nadam, self).__init__(lr, clipnorm, clipvalue, clip_alg=clip_alg,
                                name=name)
    with tf.variable_scope(self.name):
      self.iterations = _as_variable(0, name='iterations', roles=Auxiliary)
      self.m_schedule = _as_variable(1., name='m_schedule', roles=Auxiliary)
      self.schedule_decay = schedule_decay
      self.beta_1 = _as_variable(beta_1, name='beta_1',
                                 roles=OptimizerHyperParameter)
      self.beta_2 = _as_variable(beta_2, name='beta_2',
                                 roles=OptimizerHyperParameter)
      self.epsilon = epsilon
      raise NotImplementedError


class Adagrad(Optimizer):
  """Adagrad updates

  Scale learning rates by dividing with the square root of accumulated
  squared gradients. See [1]_ for further description.

  Parameters
  ----------
  lr : float or symbolic scalar
      The learning rate controlling the size of update steps
  initial_accumulator_value: A floating point value.
      Starting value for the gradients accumulators, must be positive.
  dual_avg: bool
      if True, use Adagrad Dual Averaging algorithm for sparse linear models.
  decay_steps: A scalar `int32` or `int64` `Tensor` or a Python number.
    Must be positive. (e.g. decay every 100000 steps with a base of 0.96)
  decay_rate: A scalar `float32` or `float64` `Tensor` or a
    Python number.  The decay rate.
  staircase: boolean.
      If `True` decay the learning rate at discrete intervals
  clipnorm: float >= 0. Gradients will be clipped
      when their L2 norm exceeds this value.
  clipvalue: float >= 0. Gradients will be clipped
      when their absolute value exceeds this value.
  clip_alg: str
      clipping algorithm for the gradients, one of the followings are
      accepted: "norm", "total_norm", "avg_norm"

  Returns
  -------
  OrderedDict
      A dictionary mapping each parameter to its update expression

  Notes
  -----
  Using step size eta Adagrad calculates the learning rate for feature i at
  time step t as:

  .. math:: \\eta_{t,i} = \\frac{\\eta}
     {\\sqrt{\\sum^t_{t^\\prime} g^2_{t^\\prime,i}+\\epsilon}} g_{t,i}

  as such the learning rate is monotonically decreasing.

  Epsilon is not included in the typical formula, see [2]_.

  References
  ----------
  .. [1] Duchi, J., Hazan, E., & Singer, Y. (2011):
         Adaptive subgradient methods for online learning and stochastic
         optimization. JMLR, 12:2121-2159.

  .. [2] Chris Dyer:
         Notes on AdaGrad. http://www.ark.cs.cmu.edu/cdyer/adagrad.pdf
  """

  def __init__(self, lr=0.01, initial_accumulator_value=0.1,
               dual_avg=False,
               l1_regularization=0.0, l2_regularization=0.0,
               decay_steps=None, decay_rate=0.96, staircase=True,
               clipnorm=None, clipvalue=None, clip_alg='total_norm',
               name=None):
    super(Adagrad, self).__init__(lr=lr,
        decay_steps=decay_steps, decay_rate=decay_rate, staircase=staircase,
        clipnorm=clipnorm, clipvalue=clipvalue, clip_alg=clip_alg,
        name=name)
    with tf.variable_scope(self.name):
      self.initial_accumulator_value = float(initial_accumulator_value)
      self.l1_regularization = _as_variable(l1_regularization,
          name='l1_regularization', roles=OptimizerHyperParameter)
      self.l2_regularization = _as_variable(l2_regularization,
          name='l2_regularization', roles=OptimizerHyperParameter)
      self.dual_avg = bool(dual_avg)
      # create algorithm
      if self.dual_avg:
        self._algorithm = tf.train.AdagradDAOptimizer(learning_rate=self.lr,
            global_step=self._step,
            initial_gradient_squared_accumulator_value=self.initial_accumulator_value,
            l1_regularization_strength=self.l1_regularization,
            l2_regularization_strength=self.l2_regularization)
      else:
        self._algorithm = tf.train.AdagradOptimizer(learning_rate=self.lr,
            initial_accumulator_value=self.initial_accumulator_value)
